#
# WiKicker
#
#  CGI.pm
#
#
# Copyright (C) 2003,2004 WATANABE Yoshimasa <naney@naney.org>
#
# This file is part of the WiKicker.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc., 59
# Temple Place, Suite 330, Boston, MA 02111-1307 USA
#
#
# $Id: AbstractController.pm,v 1.18 2004/08/03 12:55:52 naney Exp $
#
#

package WiKicker::CGI::AbstractController;

use 5.005;
use strict;
use CGI qw/-no_xhtml/;
use Time::HiRes qw(gettimeofday);
use WiKicker::CGI::Response;
use AutoLoader 'AUTOLOAD';
use vars qw($Run_count);
use Log::Log4perl;

BEGIN { $Run_count = 0; }

sub new {
  my ($class, %options) = @_;

  my $start_time = [ gettimeofday ];

  my $self = bless { start_time => $start_time }, $class;
  $self->{options} = \%options;
  $self->_init_logger;
  $self->_log_environment if Log::Log4perl->get_logger->is_debug;

  return $self;
}

sub run {
  my $self = shift;

  $Run_count++;
  $CGI::XHTML = 0;

  my $request = $self->_create_and_setup_request;
  return if $self->_run_hook($request);

  my ($status, $forward) = $self->_action($request);
  unless ($status->{state}) {
    $self->_handle_error($status);
    return;
  }

  my $response = $self->_create_and_setup_response($request);

  my $result = $self->_view($forward, $request, $response);
  unless ($result->{state}) {
    $self->_handle_error($result);
  }
  $self->_log($request);
  $self->_run_post_handler($request);

  #-IFDEBUG: WiKicker::total_size('$request', $request);
  #-IFDEBUG: WiKicker::total_size('$response', $response);
  #IFDEBUG: WiKicker::cycle_check($request);
  #IFDEBUG: WiKicker::cycle_check($response);
}

sub _action {
  my ($self, $request) = @_;

  my $forward;
  {
    my $page = $self->_create_page($request);
    eval {
      $forward = $page->action;
      #IFDEBUG: WiKicker::cycle_check($page);
    };
    if ($@) {
      my $exception = $@;
      Log::Log4perl->get_logger->error($exception);
      if ($exception =~ /^Unknown action/) {
        return { state => 0,
                 message => $@,
                 output => 'no' };
      }
      else {
        return { state => 0,
                 message => $@,
                 output => 'no' };
      }
    }
    #-IFDEBUG: WiKicker::cycle_check($page);
  }
  return ({state => 1}, $forward);
}

sub _view {
  my ($self, $forward, $request, $response) = @_;
  my $view_class = $request->configuration->view_class($forward,
                                                       $request->format);
  eval "use $view_class";
  if ($@) {
    return { state => 0,
             message => "Can't load view class '$view_class': $@",
             output => 'no',
           };
  }

  eval {
    #$view_class->new($request, $response)->view;
    my $view = $view_class->new($request, $response);
    $view->view;
    #IFDEBUG: WiKicker::cycle_check($view);
    $response->flush;
  };
  if ($@) {
    return { state => 0,
             message => "View error: $@",
             output => 'half',
           };
  }

  return { state => 1 };
}

sub _log {
  my ($self, $request) = @_;

  # for override;
}

sub _run_post_handler {
  my ($self, $request) = @_;

  # for override;
}

sub _create_and_setup_request {
  my $self = shift;

  my $request =
    $self->_get_request($self->_get_environment($self->_get_configuration),
                        new CGI);
  $request->attribute('run_count', $Run_count);
  $request->attribute('start_time', $self->{start_time});

  return $request;
}

sub _create_and_setup_response {
  my ($self, $request) = @_;

  my $response = WiKicker::CGI::Response->new;
  my $cookie_domain = $request->configuration->cookie_domain;
  my $cookie_path = $request->configuration->cookie_path;

  $response->add_cookie(WiKicker::CGI::Cookie::make_user_cookie($request->user, $cookie_domain, $cookie_path));

  return $response;
}

sub _get_configuration {
  my $self = shift;

  my %options = ();
  if (exists $self->{options}->{properties_file}) {
    $options{properties_file} = $self->{options}->{properties_file};
  }

  my $class = $self->_get_configuration_class;

  return $class->new(%options);
}

sub _init_logger {
  my $self = shift;

  if (exists $self->{options}->{log_conf_file}) {
    eval {
      Log::Log4perl->init_once($self->{options}->{log_conf_file});
    };
    if ($@) {
      warn $@;
    }
  }
  elsif (exists $self->{options}->{log_conf}) {
    my $conf = $self->{options}->{log_conf};
    eval {
      Log::Log4perl->init_once(\$conf);
    };
    if ($@) {
      warn $@;
    }
  }
  else {
    Log::Log4perl->easy_init($Log::Log4perl::ERROR);
  }
  Log::Log4perl->get_logger->debug('init Log::Log4perl');
}

1;

sub _handle_error;
sub _log_environment;

__END__

sub _handle_error {
  my ($self, $result) = @_;

  if ($result->{output} eq 'no') {
    my $query = CGI->new;
    print $query->header;
    print $query->start_html('Error');
    print $query->h1('Error');
    print $query->p($result->{message});
    print $query->end_html;
  }
  else {
    print $result->{message};
  }

  #warn $result->{message};
}

sub _log_environment {
  my $self = shift;

  my $logger = Log::Log4perl->get_logger;
  for my $v (sort keys %ENV) {
    $logger->debug("\$ENV{$v} = $ENV{$v}");
  }
}
