#
# WiKicker
#
#  SecondaryDatabase.pm
#
#
# Copyright (C) 2004 WATANABE Yoshimasa <naney@naney.org>
#
# This file is part of the WiKicker.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc., 59
# Temple Place, Suite 330, Boston, MA 02111-1307 USA
#
#
# $Id: SecondaryDatabase.pm,v 1.4 2004/12/13 12:53:46 naney Exp $
#
#

package WiKicker::Diki::SecondaryDatabase;

use 5.005;
use strict;
use Carp;
use File::Path;
use WiKicker::Diki::ArticleDB;
use WiKicker::Diki::TermDB;
use WiKicker::Diki::KeywordArticle;

sub new {
  my ($class, %param) = @_;

  unless (exists $param{directory}) {
    croak 'No directory';
  }

  unless (exists $param{environment}) {
    croak 'No environment';
  }

  my $self = bless { directory => $param{directory},
                     environment => $param{environment},
                   }, $class;
  $self->_initialize;

  return $self;
}

sub directory            { shift->{directory} }
sub environment          { shift->{environment} }
sub article_db_file_name { shift->{article_db_file_name} }
sub term_db_file_name    { shift->{term_db_file_name} }

sub _initialize {
  my $self = shift;

  unless (-e $self->directory) {
    eval { mkpath([ $self->directory ], 0, 0777) };
    if ($@) {
      my $message = $@;
      $message =~ s/at\s+(.*)line\s+(\d+)//;
      croak q(Can't create directory ') . $self->directory . "': $message";
    }
  }

  $self->{article_db_file_name} = $self->directory . '/article.db';
  $self->{term_db_file_name}    = $self->directory . '/term.db';
}

sub get_recent_diaries {
  my ($self, $size) = @_;

  my $diary_day_list;
  my $db = $self->_open_article_db('read');
  eval {
    $diary_day_list = $db->get_recent_diaries($size);
  };
  my $status = $@;
  undef $db;
  if ($status) {
    croak $status;
  }
  return $diary_day_list;
}

sub get_article_diaries {
  my ($self, $range) = @_;

  my $diary_day_list;
  my $db = $self->_open_article_db('read');
  eval {
    $diary_day_list = $db->get_diaries([ $range ]);
  };
  my $status = $@;
  undef $db;
  if ($status) {
    croak $status;
  }
  return $diary_day_list;
}

sub get_one_day_diaries {
  my ($self, $range) = @_;

  my $diary_day_list;
  my $db = $self->_open_article_db('read');
  eval {
    $diary_day_list = $db->get_one_day_diaries($range);
  };
  my $status = $@;
  undef $db;
  if ($status) {
    croak $status;
  }
  return $diary_day_list;
}

sub get_ten_day_diaries {
  my ($self, $range) = @_;

  my $diary_day_list;
  my $db = $self->_open_article_db('read');
  eval {
    $diary_day_list = $db->get_ten_day_diaries($range);
  };
  my $status = $@;
  undef $db;
  if ($status) {
    croak $status;
  }
  return $diary_day_list;
}

sub get_recent_keyword_diaries {
  my ($self, $name, $size) = @_;

  my $article_db = $self->_open_article_db('read');
  my $term_db = $self->_open_term_db;
  my $list = $term_db->get_recent_diaries_by_term($name, $size);
  return $article_db->get_diaries($list);
}

sub get_keywords {
  my ($self, $term) = @_;

  my $article_db = $self->_open_article_db('read');
  my $term_db = $self->_open_term_db;
  my $list = $term_db->get_keyword_articles_by_term($term);
  return $article_db->get_keywords($list);
}


sub update {
  my ($self, $diary_day_list, $update_request) = @_;

  my $article_db = $self->_open_article_db('write');
  my $term_db = $self->_open_term_db;

  for my $diff_item (@$update_request) {
    my $article_id = $diff_item->article_id;
    if ($diff_item->diff_type eq 'updated') {
      my $article = $diff_item->[2];
      my $last_modified = $diff_item->[3];

      my ($term_list, $term_list_modified) = $term_db->term_list;

      my $diary_article
        = $diary_day_list->get_article($article_id);

      $diary_article
        ->update(text => $article,
                 last_modified => $last_modified,
                 term_list => $term_list,
                 term_list_modified => $term_list_modified,
                 environment => $self->environment);
      $article_db->update_diary_article($diary_article);
      $term_db->update_diary_article_entry($diary_article);
    }
    elsif ($diff_item->diff_type eq 'new') {
      my $article = $diff_item->[2];
      my $last_modified = $diff_item->[3];

      my ($term_list, $term_list_modified) = $term_db->term_list;

      my $diary_article
        = WiKicker::Diki::DiaryArticle->new(id => $article_id,
                                            text => $article,
                                            last_modified => $last_modified,
                                            term_list => $term_list,
                                            term_list_modified => $term_list_modified,
                                           environment => $self->environment);
      $article_db->regist_diary_article($diary_article);
      $term_db->regist_diary_article($diary_article);
      $diary_day_list->add_article($diary_article);
    }
    elsif ($diff_item->diff_type eq 'deleted') {
      $article_db->remove_diary_article($article_id);
      $term_db->remove_diary_article($article_id);
      $diary_day_list->remove_article($article_id);
    }
    else {
      croak 'Unknwon diff type: ' . $diff_item->diff_type;
    }
  }
}

sub update_keyword {
  my ($self, $keyword_list, $update_request) = @_;

  my $article_db = $self->_open_article_db('write');
  my $term_db = $self->_open_term_db;

  for my $diff_item (@$update_request) {
    my $article_id = $diff_item->[0];
    if ($diff_item->[1] eq 'updated') {
      my $article = $diff_item->[2];
      my $last_modified = $diff_item->[3];

      my ($term_list, $term_list_modified) = $term_db->term_list;

      my $keyword_article
        = $keyword_list->get_article($article_id);

      $keyword_article
        ->update(text => $article,
                 last_modified => $last_modified,
                 term_list => $term_list,
                 term_list_modified => $term_list_modified,
                 environment => $self->environment);
      $article_db->update_keyword_article($keyword_article);
      $term_db->update_keyword_article($keyword_article);
    }
    elsif ($diff_item->[1] eq 'deleted') {
      $article_db->remove_keyword_article($article_id);
      $term_db->remove_keyword_article($article_id);
      $keyword_list->remove_article($article_id);
    }
  }
}

# perlfaq4
sub fisher_yates_shuffle {
  my $deck = shift;             # $deck is a reference to an array
  my $i = @$deck;
  while ($i--) {
    my $j = int rand ($i+1);
    @$deck[$i,$j] = @$deck[$j,$i];
  }
}


sub check_diary_term {
  my ($self, $diary_day_list, $max_update) = @_;

  $max_update = 1 unless defined $max_update;

  my $article_db = $self->_open_article_db('write');
  my $term_db = $self->_open_term_db;

  my $article_id_to_modified_map
    = $diary_day_list->article_term_list_modified_set;

  my $update = 0;
  my ($term_list, $term_list_modified) = $term_db->term_list;
  #for my $article_id (sort { $b cmp $a } keys %$article_id_to_modified_map) {
  my @article_id_list = keys %$article_id_to_modified_map;
  fisher_yates_shuffle(\@article_id_list);
  for my $article_id (@article_id_list) {
    if ($article_id_to_modified_map->{$article_id} < $term_list_modified) {
      my ($term_list, $term_list_modified) = $term_db->term_list;
      my $diary_article
        = $diary_day_list->get_article($article_id);

      $diary_article
        ->update(text => $diary_article->text,
                 last_modified => $diary_article->last_modified,
                 term_list => $term_list,
                 term_list_modified => $term_list_modified,
                 environment => $self->environment);
      $article_db->update_diary_article($diary_article);
      $term_db->update_diary_article_entry($diary_article);

      $update++;
      if ($update >= $max_update) {
        Log::Log4perl->get_logger->debug("update $update articles' term");
        return $update;
      }
    }
  }
  return 0;
}

sub check_keyword_term {
  my ($self, $keyword_list) = @_;

  my $article_db = $self->_open_article_db('write');
  my $term_db = $self->_open_term_db;

  my $article_id_to_modified_map
    = $keyword_list->article_term_list_modified_set;

  my ($term_list, $term_list_modified) = $term_db->term_list;
  for my $article_id (keys %$article_id_to_modified_map) {
    if ($article_id_to_modified_map->{$article_id} < $term_list_modified) {
      my ($term_list, $term_list_modified) = $term_db->term_list;
      my $keyword_article
        = $keyword_list->get_article($article_id);

      $keyword_article
        ->update(text => $keyword_article->text,
                 last_modified => $keyword_article->last_modified,
                 term_list => $term_list,
                 term_list_modified => $term_list_modified,
                 environment => $self->environment);
      $article_db->update_keyword_article($keyword_article);
      $term_db->update_keyword_article($keyword_article);
      return 1; # update only one keyword article (for response)
    }
  }
  return 0;
}


sub check_keyword_files_update {
  my ($self, $primary_db) = @_;

  my $article_db = $self->_open_article_db('read');

  my $iterator = $primary_db->keyword_id_modified_iterator;
  my %keyword_in_primary_db;
  while ($iterator->has_next) {
    my ($keyword_id, $modified) = $iterator->get_next;
    $keyword_in_primary_db{$keyword_id}++;
    my $status = $article_db
      ->check_keyword_article($keyword_id, $modified);
    if ($status eq 'new') {
      Log::Log4perl->get_logger->debug("new keyword $keyword_id");

      undef $article_db;
      $article_db = $self->_open_article_db('write');
      my $term_db = $self->_open_term_db;

      return if ($article_db->check_keyword_article($keyword_id, $modified) ne 'new'); # recheck

      my ($term_list, $term_list_modified) = $term_db->term_list;
      my $keyword_article
        = WiKicker::Diki::KeywordArticle
          ->new(id => $keyword_id,
                text => $primary_db->get_keyword_article($keyword_id),
                last_modified => $modified,
                term_list => $term_list,
                term_list_modified => $term_list_modified,
                environment => $self->environment);
      $article_db->regist_keyword_article($keyword_article);
      $term_db->regist_keyword_article($keyword_article);
      return;
    }
    if ($status eq 'updated') {
      Log::Log4perl->get_logger->debug("update keyword $keyword_id");

      undef $article_db;
      $article_db = $self->_open_article_db('write');
      my $term_db = $self->_open_term_db;

      return if ($article_db->check_keyword_article($keyword_id, $modified) ne 'updated'); # recheck


      my ($term_list, $term_list_modified) = $term_db->term_list;
      my $keyword_article
        = WiKicker::Diki::KeywordArticle
          ->new(id => $keyword_id,
                text => $primary_db->get_keyword_article($keyword_id),
                last_modified => $modified,
                term_list => $term_list,
                term_list_modified => $term_list_modified,
                environment => $self->environment);
      $article_db->update_keyword_article($keyword_article);
      $term_db->update_keyword_article($keyword_article);
      return;
    }

    next if $status eq 'not changed';

    Log::Log4perl->get_logger->debug("Unknown status: $status");
  }

  my $deleted = $article_db->get_one_deleted_keyword(\%keyword_in_primary_db);
  if ($deleted ne '') {
    Log::Log4perl->get_logger->debug("Remove keyword $deleted");

    undef $article_db;
    $article_db = $self->_open_article_db('write');
    my $term_db = $self->_open_term_db;

    $article_db->remove_keyword_article($deleted);
    $term_db->remove_keyword_article($deleted);
  }
}

sub term_list {
  my $self = shift;

  my $article_db = $self->_open_article_db('read');
  my $term_db = $self->_open_term_db;

  my ($term_list, $modified) = $term_db->term_list;

  return $term_list;
}

sub _open_article_db {
  my ($self, $mode) = @_;

  return WiKicker::Diki::ArticleDB->new(file_name => $self->article_db_file_name,
                                        mode => $mode);
}

sub _open_term_db {
  my $self = shift;

  return WiKicker::Diki::TermDB->new(file_name => $self->term_db_file_name);
}

1;
