#
# WiKicker
#
#  TermDB.pm
#
#
# Copyright (C) 2004 WATANABE Yoshimasa <naney@naney.org>
#
# This file is part of the WiKicker.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc., 59
# Temple Place, Suite 330, Boston, MA 02111-1307 USA
#
#
# $Id: TermDB.pm,v 1.2 2004/08/03 12:55:52 naney Exp $
#
#

package WiKicker::Diki::TermDB;

use 5.005;
use strict;
use Carp;
use DB_File;
use Storable qw(freeze thaw);
use Fcntl qw(O_RDWR O_CREAT);

#IFDEBUG: use Carp::Assert;

# t@size           | num
# t@mod            : time
# t:<term>         | TermRecord
# d:<diary id>     | DiaryRecord
# k:<keyword id>   | KeywordRecord
# T#<term index>   | string
# D#<diary index>  | string
# K#<keyword index>| string
# T@next           | num
# D@next           | num
# K@next           | num

# DiaryRecord (WiKicker::Diki::TermDB::DiaryRecord)
# [
#  diary index,
#  {explicit_term_index_1 => 1, explicit_term_index_2 => 1, ...}
#  {implicit_term_index_1 => 1, implicit_term_index_2 => 1, ...}
# ]

# KeywordRecord
# [
#  keyword index,
#  {explicit_term_index_1 => 1, explicit_term_index_2 => 1, ...}
#  {implicit_term_index_1 => 1, implicit_term_index_2 => 1, ...}
#  {defined_term_index_1  => 1, defined_term_index_2  => 1, ...}
# ]

# TermRecord
# [
#  term index,
#  {explicit_diary_index_1 => 1, explicit_diary_index_2 => 1, ...}
#  {implicit_diary_index_1 => 1, implicit_diary_index_2 => 1, ...}
#  {explicit_keyword_index_1 => 1, explicit_keyword_index_2 => 1, ...}
#  {implicit_keyword_index_1 => 1, implicit_keyword_index_2 => 1, ...}
#  {defined_keyword_index_1 => 1, defined_keyword_index_2 => 1, ...}
# ]

sub new {
  my ($class, %param) = @_;

  unless (exists $param{file_name}) {
    croak 'No file_name';
  }

  my $self = bless { file_name => $param{file_name} }, $class;
  $self->_initialize;

  return $self;
}

sub DESTROY {
  my $self = shift;

  $self->_close_db;
}

sub _initialize {
  my $self = shift;

  $self->_open_db;
}

sub _open_db {
  my $self = shift;

  my %db_hash;
  my $db_obj = tie %db_hash, 'DB_File',
    $self->{file_name}, O_RDWR|O_CREAT, 0666, $DB_BTREE;
  if ($db_obj) {
    Log::Log4perl->get_logger->debug('Open TermDB');
    $self->{db_obj} = $db_obj;
    $self->{db_hash} = \%db_hash;
    return;
  }
  else {
    croak q(Can't open database: ) . $self->{file_name} . ": $!";
  }
}

sub hash { shift->{db_hash} }

sub _close_db {
  my $self = shift;

  $self->{db_obj}->sync;
  undef $self->{db_obj};
  untie %{$self->{db_hash}};
  delete $self->{db_hash};
  Log::Log4perl->get_logger->debug('Close TermDB');
}


sub term_list {
  my $self = shift;

  if (exists $self->hash->{'t@size'}) {
    return $self->_get_term_list_with_size($self->hash->{'t@size'});
  }

  my @term_list = ();
  my $size = $self->_get_term_list_from_hash(\@term_list);
  Log::Log4perl->get_logger->debug("Num of term: $size");
  $self->hash->{'t@size'} = $size;

  return (\@term_list, $self->_get_term_set_modified);
}

sub _get_term_list_with_size {
  my ($self, $size) = @_;

  my @term_list = ();
  if ($size > 0) {
    $#term_list = $size - 1;
  }
  my $new_size = $self->_get_term_list_from_hash(\@term_list);
  if ($new_size != $size) {
    Log::Log4perl->get_logger->warn('t@size is ' . $size . ', but ' . $new_size);
    $self->hash->{'t@size'} = $new_size;
    $#term_list = $new_size - 1;
  }
  #IFDEBUG: assert(scalar(@term_list) == $new_size);
  return (\@term_list, $self->_get_term_set_modified);
}

sub _get_term_list_from_hash {
  my ($self, $term_list_ref) = @_;

  my $i = 0;
  # @TODO don't foreach all key
  foreach (keys %{$self->hash}) {
    my $key = $_;
    if (substr($key, 0, 2) eq 't:') {
      #IFDEBUG: assert(length($key) > 3);
      $term_list_ref->[$i] = substr($key, 2);
      $i++;
    }
  }
  Log::Log4perl->get_logger->debug("Term size (in DB): $i");
  return $i;
}

sub _exists_term {
  my ($self, $term) = @_;

  return exists $self->hash->{"t:$term"};
}

sub _exists_term_index {
  my ($self, $term_index) = @_;

  return exists $self->hash->{"T#$term_index"};
}

sub _exists_keyword_index {
  my ($self, $keyword_index) = @_;

  return exists $self->hash->{"K#$keyword_index"};
}

sub _term_index_to_term {
  my ($self, $term_index) = @_;

  #IFDEBUG: affirm {
  #IFDEBUG:   warn "unknown term_index $term_index"
  #IFDEBUG:     unless exists $self->hash->{'T#' . $term_index};
  #IFDEBUG:   exists $self->hash->{'T#' . $term_index};
  #IFDEBUG: };
  return $self->hash->{'T#' . $term_index};
}

sub _get_term_record {
  my ($self, $term) = @_;

  #IFDEBUG: affirm { exists $self->hash->{"t:$term"} };
  #IFDEBUG: affirm { defined $self->hash->{"t:$term"} };
  return WiKicker::Diki::TermDB::TermRecord->new(serialized =>
       $self->hash->{"t:$term"});
}

sub _set_term_record {
  my ($self, $term, $record) = @_;

  #IFDEBUG: affirm { $term ne '' };
  #IFDEBUG: affirm { my $i = $record->index;
  #IFDEBUG:          exists $self->hash->{'T#' . $i} &&
  #IFDEBUG:            $self->hash->{'T#' . $i} eq $term;
  #IFDEBUG: };
  $self->hash->{"t:$term"} = $record->serialize;
}

sub _new_term_record {
  my ($self, $term) = @_;

  my $new_index = $self->_get_new_term_index($term);
  return WiKicker::Diki::TermDB::TermRecord->new(index => $new_index);
}

sub _new_diary_record {
  my ($self, $diary_id) = @_;

  my $index = $self->_get_new_diary_index($diary_id);
  return WiKicker::Diki::TermDB::DiaryRecord->new(index => $index);
}

sub _new_keyword_record {
  my ($self, $keyword_id) = @_;

  my $index = $self->_get_new_keyword_index($keyword_id);
  return WiKicker::Diki::TermDB::KeywordRecord->new(index => $index);
}


sub _get_diary_record {
  my ($self, $diary_id) = @_;
  return WiKicker::Diki::TermDB::DiaryRecord->new(serialized => $self->hash->{'d:' . $diary_id});
}

sub _set_diary_record {
  my ($self, $diary_id, $record) = @_;

  $self->hash->{'d:' . $diary_id} = $record->serialize;
}

sub _set_keyword_record {
  my ($self, $keyword_id, $record) = @_;

  #IFDEBUG: affirm { $self->hash->{'K#' . $record->index} eq $keyword_id }
  $self->hash->{'k:' . $keyword_id} = $record->serialize;
}

sub _get_keyword_record {
  my ($self, $keyword_id) = @_;
  return WiKicker::Diki::TermDB::KeywordRecord->new(serialized => $self->hash->{'k:' . $keyword_id});
}

sub _get_diary_id_from_index {
  my ($self, $index) = @_;

  #IFDEBUG: affirm { exists $self->hash->{'D#' . $index} };
  return $self->hash->{'D#' . $index};
}

sub _get_keyword_id_from_index {
  my ($self, $index) = @_;

  #IFDEBUG: affirm { exists $self->hash->{'K#' . $index} };
  return $self->hash->{"K#$index"};
}

sub get_recent_diaries_by_term {
  my ($self, $term, $size) = @_;

  if ($self->_exists_term($term)) {
    my $record = $self->_get_term_record($term);
    my $refered_diary_index_list = $record->refered_diary_index_list;
    my @result = ();
    for my $index (@$refered_diary_index_list) {
      push(@result, $self->_get_diary_id_from_index($index));
    }
    return \@result;
  }

  return [];
}

sub get_keyword_articles_by_term {
  my ($self, $term) = @_;

  if ($self->_exists_term($term)) {
    my $record = $self->_get_term_record($term);
    my $keyword_index_list = $record->keyword_define_index_list;
    my @result = ();
    for my $index (@$keyword_index_list) {
      if ($self->_exists_keyword_index($index)) {
        push(@result, $self->_get_keyword_id_from_index($index));
      }
      else {
        Log::Log4perl->get_logger->error("no keyword K#$index in term T#" . $record->index);
      }
    }
    Log::Log4perl->get_logger
        ->debug(sub { '[T#' . $record->index . '] : ' . join(' ', map {"[k:$_]"} @result); });
    return \@result;
  }

  return [];
}

sub regist_diary_article {
  my ($self, $diary_article) = @_;

  my $diary_record = $self->_new_diary_record($diary_article->id);

  my $term_size_delta = 0;
  for my $term (@{$diary_article->explicit_term_list}) {
    next if $term eq '';
    my ($term_index, $new) = $self->_set_diary_term($term,
                                                    $diary_record->index,
                                                    'explicit');
    $diary_record->set_explicit_term($term_index);
    $term_size_delta += $new;
  }

  for my $term (@{$diary_article->implicit_term_list}) {
    my ($term_index, $new) = $self->_set_diary_term($term,
                                                    $diary_record->index,
                                                    'implicit');
    $diary_record->set_implicit_term($term_index);
    $term_size_delta += $new;
  }

  $self->_term_size_delta_update($term_size_delta);

  $self->_set_diary_record($diary_article->id, $diary_record);

  return;
}

sub remove_diary_article {
  my ($self, $diary_article_id) = @_;

  my $term_size_delta = 0;

  my $diary_record = $self->_get_diary_record($diary_article_id);
  my $diary_index = $diary_record->index;
  #IFDEBUG:assert($self->hash->{"D#$diary_index"} eq $diary_article_id);
  Log::Log4perl->get_logger->debug("Remove d:$diary_article_id");
  Log::Log4perl->get_logger->debug("Remove D#$diary_index");

  # remove implicit term index
  for my $term_index (@{$diary_record->get_implicit_term_index_list}) {
    $self->_remove_diary_implicit_term($term_index, $diary_index);
  }

  # remove explicit term index
  my @explicit_term_index_list = @{$diary_record->get_explicit_term_index_list};
  for my $term_index (@explicit_term_index_list) {
    $self->_remove_diary_explicit_term($term_index, $diary_index);
  }

  for my $explicit_term_index (@explicit_term_index_list) {
    $term_size_delta -=
      $self->_remove_term_if_no_definition($explicit_term_index);
  }

  delete $self->hash->{'d:' . $diary_article_id};
  delete $self->hash->{'D#' . $diary_index};

  $self->_term_size_delta_update($term_size_delta);
}

sub remove_keyword_article {
  my ($self, $keyword_id) = @_;

  my $term_size_delta = 0;

  my $keyword_record = $self->_get_keyword_record($keyword_id);
  my $keyword_index = $keyword_record->index;
  #IFDEBUG:assert($self->hash->{"K#$keyword_index"} eq $keyword_id);
  Log::Log4perl->get_logger->debug("Remove k:$keyword_id");
  Log::Log4perl->get_logger->debug("Remove K#$keyword_index");

  # remove implicit term index
  for my $term_index (@{$keyword_record->get_implicit_term_index_list}) {
    $self->_remove_keyword_implicit_term($term_index, $keyword_index);
  }

  # remove explicit term index
  my @explicit_term_index_list = @{$keyword_record
                                     ->get_explicit_term_index_list};
  for my $term_index (@explicit_term_index_list) {
    $self->_remove_keyword_explicit_term($term_index, $keyword_index);
  }
  # remove
  my @defined_term_index_list = @{$keyword_record
                                    ->get_defined_term_index_list};

  for my $term_index (@defined_term_index_list) {
    $self->_remove_keyword_defined_term($term_index, $keyword_index);
  }

  my %term_map;
  map { $term_map{$_}++ } @explicit_term_index_list;
  map { $term_map{$_}++ } @defined_term_index_list;


  for my $term_index (keys %term_map) {
    $term_size_delta -=
      $self->_remove_term_if_no_definition($term_index);
  }

  delete $self->hash->{'k:' . $keyword_id};
  delete $self->hash->{'K#' . $keyword_index};

  $self->_term_size_delta_update($term_size_delta);
}

sub _remove_keyword_explicit_term {
  my ($self, $term_index, $keyword_index) = @_;

  my $term = $self->_term_index_to_term($term_index);
  my $term_record = $self->_get_term_record($term);

  $term_record->remove_keyword($keyword_index, 'explicit');
  $self->_set_term_record($term, $term_record);
}

sub _remove_keyword_implicit_term {
  my ($self, $term_index, $keyword_index) = @_;

  if ($self->_exists_term_index($term_index)) {
    my $term = $self->_term_index_to_term($term_index);
    my $term_record = $self->_get_term_record($term);

    $term_record->remove_keyword($keyword_index, 'implicit');
    $self->_set_term_record($term, $term_record);
  }
  else {
    Log::Log4perl->get_logger->error("no T#$term_index");
  }
}

sub _remove_keyword_defined_term {
  my ($self, $term_index, $keyword_index) = @_;

  my $term = $self->_term_index_to_term($term_index);
  my $term_record = $self->_get_term_record($term);

  $term_record->remove_keyword($keyword_index, 'define');
  $self->_set_term_record($term, $term_record);
}

sub _remove_term_if_no_definition {
  my ($self, $term_index) = @_;

  my $term = $self->_term_index_to_term($term_index);
  my $term_record = $self->_get_term_record($term);

  if ($term_record->no_definition) {
    Log::Log4perl->get_logger->debug("T#$term_index has no definition");
    my @implicit_diaries = $term_record->implicit_diaries;
    my @implicit_keywords = $term_record->implicit_keywords;

    for my $diary_index (@implicit_diaries) {
      my $diary_id = $self->_get_diary_id_from_index($diary_index);
      my $diary_record = $self->_get_diary_record($diary_id);
      $diary_record->remove_implicit_term($term_index);
      $self->_set_diary_record($diary_id, $diary_record);
      Log::Log4perl
          ->get_logger->debug("[d:$diary_id] Remove implicit T#$term_index");
    }

    for my $keyword_index (@implicit_keywords) {
      my $keyword_id = $self->_get_keyword_id_from_index($keyword_index);
      my $keyword_record = $self->_get_keyword_record($keyword_id);
      $keyword_record->remove_implicit_term($term_index);
      $self->_set_keyword_record($keyword_id, $keyword_record);
      Log::Log4perl
          ->get_logger->debug("[k:$keyword_id] Remove T#$term_index");
    }

    Log::Log4perl->get_logger->debug("[T#$term_index / t:$term] Remove entry");
    delete $self->hash->{'t:' . $term};
    delete $self->hash->{'T#' . $term_index};

    return 1;
  }
  else {
    return 0;
  }
}

sub _remove_diary_explicit_term {
  my ($self, $term_index, $diary_index) = @_;

  my $term = $self->_term_index_to_term($term_index);
  my $term_record = $self->_get_term_record($term);

  $term_record->remove_diary('explicit', $diary_index);
  $self->_set_term_record($term, $term_record);
}

sub _remove_diary_implicit_term {
  my ($self, $term_index, $diary_index) = @_;

  unless ($self->_exists_term_index($term_index)) {
    Log::Log4perl->get_logger->error("No Term T#$term_index");
    return;
  }

  my $term_id = $self->_term_index_to_term($term_index);
  my $term_record = $self->_get_term_record($term_id);
  $term_record->remove_diary('implicit', $diary_index);
  $self->_set_term_record($term_id, $term_record);

  return;
}

sub regist_keyword_article {
  my ($self, $keyword_article) = @_;

  my $keyword_article_id = $keyword_article->id;
  my $keyword_record = $self->_new_keyword_record($keyword_article_id);
  my $keyword_index = $keyword_record->index;
  Log::Log4perl->get_logger->debug("New keyword record: k:$keyword_article_id = K#$keyword_index");

  my $term_size_delta = 0;
  for my $term (@{$keyword_article->explicit_term_list}) {
    my ($term_index, $new) = $self->_set_keyword_term($term,
                                                      $keyword_index,
                                                      'explicit');
    $keyword_record->set_explicit_term($term_index);
    $term_size_delta += $new;
  }

  for my $term (@{$keyword_article->implicit_term_list}) {
    my ($term_index, $new) = $self->_set_keyword_term($term,
                                                      $keyword_index,
                                                      'implicit');
    $keyword_record->set_implicit_term($term_index);
    $term_size_delta += $new;
  }

  for my $term (@{$keyword_article->defined_term_list}) {
    my ($term_index, $new) = $self->_set_keyword_term($term,
                                                      $keyword_index,
                                                      'define');
    $keyword_record->set_defined_term($term_index);
    $term_size_delta += $new;
  }

  $self->_term_size_delta_update($term_size_delta);

  $self->_set_keyword_record($keyword_article->id, $keyword_record);

  #IFDEBUG: affirm { $self->_exists_keyword_index($keyword_index) }

  return;
}

sub _term_size_delta_update {
  my ($self, $term_size_delta) = @_;

  if ($term_size_delta != 0) {
    Log::Log4perl->get_logger->debug("delta of term: $term_size_delta");
    $self->hash->{'t@size'} += $term_size_delta;
    $self->_set_term_set_modified(time);
  }
}


sub update_diary_article_entry {
  my ($self, $diary_article) = @_;

  my $new_explicit_term_list = $diary_article->explicit_term_list;
  my $new_implicit_term_list = $diary_article->implicit_term_list;
  my $diary_record = $self->_get_diary_record($diary_article->id);
  my $diary_index = $diary_record->index;
  my $diary_id = $diary_article->id;
  my $old_explicit_term_set =
    $self->make_term_to_term_index_map($diary_record->get_explicit_term_index_list);
  my $old_implicit_term_set =
    $self->make_term_to_term_index_map($diary_record->get_implicit_term_index_list);

  # added
  my $term_set_changed = 0;
  my $term_size_delta = 0;
  for my $new_term (@$new_explicit_term_list) {
    unless (exists $old_explicit_term_set->{$new_term}) {
      my ($term_index, $new) = $self->_set_diary_term($new_term,
                                                      $diary_index,
                                                      'explicit'
                                                     );
      $diary_record->set_explicit_term($term_index);
      Log::Log4perl->get_logger->debug("[d:$diary_id] Add explicit T#$term_index" . ($new ? ' (new)' : ''));
      $term_size_delta += $new;
      $term_set_changed |= $new;
    }
    delete $old_explicit_term_set->{$new_term};
  }

  for my $new_term (@$new_implicit_term_list) {
    unless (exists $old_implicit_term_set->{$new_term}) {
      my ($term_index, $new) = $self->_set_diary_term($new_term,
                                                      $diary_index,
                                                      'implicit');
      $diary_record->set_implicit_term($term_index);
      Log::Log4perl->get_logger->debug("[d:$diary_id] Add implicit T#$term_index" . ($new ? ' (new)' : ''));
      $term_size_delta += $new;
      $term_set_changed |= $new;
    }
    delete $old_implicit_term_set->{$new_term};
  }

  # removed?
  for my $old_term (keys %$old_implicit_term_set) {
    Log::Log4perl->get_logger->debug("[d:$diary_id] Remove implicit T:$old_term");
    my $old_term_index = $old_implicit_term_set->{$old_term};
    $diary_record->remove_implicit_term($old_term_index);
    $self->_remove_diary_implicit_term($old_term_index, $diary_index);
  }

  for my $old_term (keys %$old_explicit_term_set) {
    my $old_term_index = $old_explicit_term_set->{$old_term};
    Log::Log4perl->get_logger->debug("[d:$diary_id] Remove explicit T#$old_term_index");
    $diary_record->remove_explicit_term($old_term_index);

    $self->_remove_diary_explicit_term($old_term_index, $diary_index);
  }

  # write here
  $self->_set_diary_record($diary_article->id, $diary_record);

  # this loop may change diary record
  for my $old_term (keys %$old_explicit_term_set) {
    my $old_term_index = $old_explicit_term_set->{$old_term};
    my $deleted = $self->_remove_term_if_no_definition($old_term_index);
    if ($deleted) {
      $term_set_changed = 1;
      $term_size_delta -= $deleted;
    }
  }

  if ($term_set_changed) {
    $self->_term_size_delta_update($term_size_delta);
    $self->_set_term_set_modified(time);
  }
}

sub update_keyword_article {
  my ($self, $keyword_article) = @_;

  my $new_explicit_term_list = $keyword_article->explicit_term_list;
  my $new_implicit_term_list = $keyword_article->implicit_term_list;
  my $new_defined_term_list = $keyword_article->defined_term_list;
  my $keyword_record = $self->_get_keyword_record($keyword_article->id);
  my $keyword_index = $keyword_record->index;
  my $old_explicit_term_set =
    $self->make_term_to_term_index_map($keyword_record->get_explicit_term_index_list);
  my $old_implicit_term_set =
    $self->make_term_to_term_index_map($keyword_record->get_implicit_term_index_list);
  my $old_defined_term_set =
    $self->make_term_to_term_index_map($keyword_record->get_defined_term_index_list);

  # added
  my $term_set_changed = 0;
  my $term_size_delta = 0;
  for my $new_term (@$new_explicit_term_list) {
    unless (exists $old_explicit_term_set->{$new_term}) {
      my ($term_index, $new) = $self->_set_keyword_term($new_term,
                                                        $keyword_index,
                                                        'explicit'
                                                       );
      $keyword_record->set_explicit_term($term_index);
      Log::Log4perl->get_logger->debug("Add explicit T#$term_index to " . $keyword_article->id . ' ' . ($new ? '(new)' : ''));

      $term_size_delta += $new;
      $term_set_changed |= $new;
    }
    delete $old_explicit_term_set->{$new_term};
  }
  for my $new_term (@$new_implicit_term_list) {
    unless (exists $old_implicit_term_set->{$new_term}) {
      my ($term_index, $new) = $self->_set_keyword_term($new_term,
                                                        $keyword_index,
                                                        'implicit');
      $keyword_record->set_implicit_term($term_index);
      Log::Log4perl->get_logger->debug("Add implicit T#$term_index to " . $keyword_article->id . ' ' . ($new ? '(new)' : ''));
      $term_size_delta += $new;
      $term_set_changed |= $new;
    }
    delete $old_implicit_term_set->{$new_term};
  }

  for my $new_term (@$new_defined_term_list) {
    unless (exists $old_defined_term_set->{$new_term}) {
      my ($term_index, $new) = $self->_set_keyword_term($new_term,
                                                        $keyword_index,
                                                        'define');
      $keyword_record->set_defined_term($term_index);
      Log::Log4perl->get_logger->debug("[K#$keyword_index] Add define T#$term_index" . ($new ? ' (new)' : ''));
      $term_size_delta += $new;
      $term_set_changed |= $new;
    }
    delete $old_defined_term_set->{$new_term};
  }

  # removed?
  for my $old_term (keys %$old_implicit_term_set) {
    Log::Log4perl->get_logger->debug("Remove implicit T:$old_term from " . $keyword_article->id);

    $keyword_record->remove_implicit_term($old_implicit_term_set->{$old_term});
    my $old_term_index = $old_implicit_term_set->{$old_term};
    $self->_remove_keyword_implicit_term($old_term_index,
                                         $keyword_index);
  }

  my %delete_candidate;

  for my $old_term (keys %$old_explicit_term_set) {
    Log::Log4perl->get_logger->debug("Remove explicit T:$old_term from " . $keyword_article->id);

    $keyword_record->remove_explicit_term($old_explicit_term_set->{$old_term});
    my $old_term_index = $old_explicit_term_set->{$old_term};
    $self->_remove_keyword_explicit_term($old_term_index,
                                         $keyword_index);
    $delete_candidate{$old_term_index}++;
  }

  for my $old_term (keys %$old_defined_term_set) {
    my $old_term_index = $old_defined_term_set->{$old_term};
    Log::Log4perl->get_logger->debug("[K#$keyword_index] Remove define T#$old_term_index");
    $keyword_record->remove_defined_term($old_defined_term_set->{$old_term});
    $self->_remove_keyword_defined_term($old_term_index, $keyword_index);
    $delete_candidate{$old_term_index}++;
  }

  # write here
  $self->_set_keyword_record($keyword_article->id, $keyword_record);

  for my $explicit_term_index (keys %delete_candidate) {
    if ($self->_exists_term_index($explicit_term_index)) {
      my $delta = $self->_remove_term_if_no_definition($explicit_term_index);
      $term_size_delta -= $delta;
      if ($delta) {
        $term_set_changed = 1;
      }
    }
  }


  if ($term_set_changed) {
    $self->hash->{'t@size'} += $term_size_delta;
    $self->_set_term_set_modified(time);
  }
}


sub make_term_to_term_index_map {
  my ($self, $term_index_list) = @_;

  my %result;
  for my $term_index (@$term_index_list) {
    #IFDEBUG: affirm { defined $term_index && $term_index };
    if ($self->_exists_term_index($term_index)) {
      my $term = $self->_term_index_to_term($term_index);
      $result{$term} = $self->_get_term_record($term)->index;
    }
    else {
      Log::Log4perl->get_logger->error("No term #$term_index");
    }
  }
  return \%result;
}

sub _set_diary_term {
  my ($self, $term, $diary_index, $type) = @_;

  #IFDEBUG: affirm { $term ne '' }
  my $new = 0;
  my $term_record;
  if ($self->_exists_term($term)) {
    $term_record = $self->_get_term_record($term);
  }
  else {
    $new = 1;
    $term_record = $self->_new_term_record($term);
  }
  $term_record->set_diary($type, $diary_index);
  $self->_set_term_record($term, $term_record);

  return ($term_record->index, $new);
}

sub _set_keyword_term {
  my ($self, $term, $keyword_index, $type) = @_;

  #IFDEBUG: affirm { $term ne '' }
  my $new = 0;
  my $term_record;
  if ($self->_exists_term($term)) {
    $term_record = $self->_get_term_record($term);
  }
  else {
    $new = 1;
    $term_record = $self->_new_term_record($term);
  }
  $term_record->set_keyword($type, $keyword_index);
  $self->_set_term_record($term, $term_record);

  return ($term_record->index, $new);
}


sub _get_new_diary_index {
  my ($self, $id) = @_;

  return $self->_get_new_index('D', $id);
}

sub _get_new_keyword_index {
  my ($self, $id) = @_;

  return $self->_get_new_index('K', $id);
}

sub _get_new_term_index {
  my ($self, $term) = @_;

  return $self->_get_new_index('T', $term);
}

sub _get_new_index {
  my ($self, $type, $id) = @_;

  my $index = 1;
  if (exists $self->hash->{$type . '@next'}) {
    $index = $self->hash->{$type . '@next'};
  }
  $self->hash->{$type . '@next'} = $index + 1;
  $self->hash->{$type . '#' . $index} = $id;

  return $index;
}


sub _get_term_set_modified {
  my $self = shift;

  if (exists $self->{'t@mod'}) {
    return $self->{'t@mod'};
  }

  if (exists $self->hash->{'t@mod'}) {
    return $self->{'t@mod'}
      = $self->hash->{'t@mod'};
  }

  $self->_set_term_set_modified(time);

  return $self->{'t@mod'};
}

sub _set_term_set_modified {
  my ($self, $time) = @_;

  $self->hash->{'t@mod'} = $self->{'t@mod'} = $time;
}

package WiKicker::Diki::TermDB::DiaryRecord;
use strict;
use Storable qw(freeze thaw);
use Carp;
#IFDEBUG: use Carp::Assert;

sub new {
  my ($class, %param) = @_;

  if (exists $param{serialized}) {
    my $self = thaw($param{serialized});
    #IFDUBUG: assert(ref($self) eq 'ARRAY');
    bless $self, $class;
    return $self;
  }

  else {
    croak 'no index' unless $param{index};
  }

  my $self = bless [ $param{index}, {}, {} ], $class;

  return $self;
}

sub serialize {
  my $self = shift;

  return freeze($self);
}

sub index { shift->[0] }

sub set_explicit_term {
  my ($self, $term_index) = @_;

  $self->[1]->{$term_index} = 1;
}

sub set_implicit_term {
  my ($self, $term_index) = @_;

  $self->[2]->{$term_index} = 1;
}

sub remove_explicit_term {
  my ($self, $term_index) = @_;

  delete $self->[1]->{$term_index};
}

sub remove_implicit_term {
  my ($self, $term_index) = @_;

  delete $self->[2]->{$term_index};
}

sub get_explicit_term_index_list {
  my $self = shift;

  my @result = keys %{$self->[1]};
  return \@result;
}

sub get_implicit_term_index_list {
  my $self = shift;

  my @result = keys %{$self->[2]};
  return \@result;
}

package WiKicker::Diki::TermDB::KeywordRecord;
use strict;
use Storable qw(freeze thaw);
use Carp;

sub new {
  my ($class, %param) = @_;

  if (exists $param{serialized}) {
    my $self = thaw($param{serialized});
    bless $self, $class;
    return $self;
  }
  else {
    croak 'no index' unless $param{index};
  }

  my $self = bless [ $param{index}, {}, {} ,{} ], $class;

  return $self;
}

sub serialize {
  my $self = shift;

  return freeze($self);
}


sub index { shift->[0] }

sub set_explicit_term {
  my ($self, $term_index) = @_;

  $self->[1]->{$term_index} = 1;
}

sub set_implicit_term {
  my ($self, $term_index) = @_;

  $self->[2]->{$term_index} = 1;
}

sub remove_explicit_term {
  my ($self, $term_index) = @_;

  delete $self->[1]->{$term_index};
}

sub remove_implicit_term {
  my ($self, $term_index) = @_;

  delete $self->[2]->{$term_index};
}
sub remove_defined_term {

  my ($self, $term_index) = @_;

  delete $self->[3]->{$term_index};
}

sub set_defined_term {
  my ($self, $term_index) = @_;

  $self->[3]->{$term_index} = 1;
}

sub get_explicit_term_index_list {
  my $self = shift;

  my @result = keys %{$self->[1]};
  return \@result;
}

sub get_implicit_term_index_list {
  my $self = shift;

  my @result = keys %{$self->[2]};
  return \@result;
}

sub get_defined_term_index_list {
  my $self = shift;

  my @result = keys %{$self->[3]};
  return \@result;
}

package WiKicker::Diki::TermDB::TermRecord;
use strict;
use Storable qw(freeze thaw);
use Carp;

sub new {
  my ($class, %param) = @_;

  if (exists $param{serialized}) {
    my $self = thaw($param{serialized});
    bless $self, $class;
    return $self;
  }
  else {
    croak 'no index' unless $param{index};
  }

  my $self = bless [ $param{index}, {}, {}, {}, {}, {} ], $class;

  return $self;
}

sub serialize {
  my $self = shift;

  return freeze($self);
}

sub index { shift->[0] }

sub no_definition {
  my $self = shift;

  my @def_element = (1, 3, 5);
  for my $i (@def_element) {
    return 0 if (keys %{$self->[$i]} > 0);
  }
  return 1;
}

sub set_diary {
  my ($self, $type, $diary_index) = @_;

  if ($type eq 'explicit') {
    $self->[1]->{$diary_index} = 1;
  }
  elsif ($type eq 'implicit') {
    $self->[2]->{$diary_index} = 1;
  }
  else {
    die;
  }
}

sub implicit_diaries {
  my $self = shift;

  return keys %{$self->[2]};
}

sub implicit_keywords {
  my $self = shift;

  return keys %{$self->[4]};
}

sub remove_diary {
  my ($self, $type, $diary_index) = @_;

  if ($type eq 'explicit') {
    delete $self->[1]->{$diary_index};
  }
  elsif ($type eq 'implicit') {
    delete $self->[2]->{$diary_index};
  }
  else {
    die;
  }
}

sub set_keyword {
  my ($self, $type, $keyword_index) = @_;

  if ($type eq 'explicit') {
    $self->[3]->{$keyword_index} = 1;
  }
  elsif ($type eq 'implicit') {
    $self->[4]->{$keyword_index} = 1;
  }
  elsif ($type eq 'define') {
    $self->[5]->{$keyword_index} = 1;
  }
  else {
    die;
  }
}

sub remove_keyword {
  my ($self, $keyword_index, $type) = @_;

  if ($type eq 'explicit') {
    delete $self->[3]->{$keyword_index};
  }
  elsif ($type eq 'implicit') {
    delete $self->[4]->{$keyword_index};
  }
  elsif ($type eq 'define') {
    delete $self->[5]->{$keyword_index};
  }
  else {
    confess "Unknown type $type";
  }
}

sub refered_diary_index_list {
  my $self = shift;

  my %result = ();
  for my $index (keys %{$self->[1]}) {
    $result{$index}++;
  }
  for my $index (keys %{$self->[2]}) {
    $result{$index}++;
  }
  return [ keys %result ];
}

sub keyword_define_index_list {
  my $self = shift;

  my @result = keys %{$self->[5]};
  return \@result;
}



1;
