#
# WiKicker
#
#  Properties.pm
#
#
# Copyright (C) 2003,2004 WATANABE Yoshimasa <naney@naney.org>
#
# This file is part of the WiKicker.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc., 59
# Temple Place, Suite 330, Boston, MA 02111-1307 USA
#
#
# $Id: Properties.pm,v 1.6 2004/02/15 05:06:19 naney Exp $
#
#

package WiKicker::Properties;

use 5.005;
use strict;
use FileHandle;

sub new {
  my ($class, %options) = @_;

  my $self = bless { parent => undef,
                     last_modified => 0,
                     properties_name => '(memory)',
                   }, $class;

  if (exists $options{parent}) {
    $self->{parent} = $options{parent};
  }

  if (exists $options{file_name}) {
    $self->_load_file($options{file_name});
    $self->{properties_name} = $options{file_name};
  }

  return $self;
}

sub parent {
  my ($self, $value) = @_;

  if (defined $value) {
    $self->{parent} = $value;
  }

  return $self->{parent};
}

sub get {
  my ($self, $key) = @_;

  if (exists $self->{property}->{$key}) {
    return $self->{property}->{$key};
  }
  elsif ($self->{parent}) {
    $self->{parent}->get($key);
  }
  else {
    return undef;
  }
}

sub properties_name_list {
  my $self = shift;

  my $result = [];
  if ($self->{parent}) {
    $result = $self->{parent}->properties_name_list;
  }
  unshift(@$result, $self->properties_name);
  return $result;
}

sub properties_name {
  my $self = shift;

  return $self->{properties_name};
}

sub set {
  my ($self, $key, $value) = @_;

  $self->{property}->{$key} = $value;
}

sub self_last_modified {
  my $self = shift;
  return $self->{last_modified};
  # TODO: update if call set method.
}

sub set_last_modified {
  my ($self, $time) = @_;

  $self->{last_modified} = $time;
}


sub last_modified {
  my $self = shift;

  if ($self->{parent}) {
    my $parent_last_modified = $self->{parent}->last_modified;
    my $self_last_modified = $self->self_last_modified;
    if ($parent_last_modified > $self_last_modified) {
      return $parent_last_modified;
    }
    else {
      return $self_last_modified;
    }
  }
  return $self->self_last_modified;
}

sub load {
  my ($self, $handle) = @_;

  $self->{handle} = $handle;
  $self->_load;
  delete $self->{handle};
}

sub load_file {
  my ($self, $file_name) = @_;

  $self->_load_file($file_name);
}

sub _get_line {
  my $self = shift;

  if (exists $self->{handle}) {
    my $handle = $self->{handle};
    my $line = <$handle>;
    return $line;
  }
  else {
    my $line = shift(@{$self->{text_list}});
    return $line;
  }
}

sub _load_from_text_list {
  my ($self, @text_list) = @_;

  delete $self->{handle};
  $self->{text_list} = \@text_list;
  $self->_load;
  delete $self->{text_list};
}

sub _load {
  my $self = shift;
  my $read_line;
  my $continue = 0;
  my $line;
  while (defined($read_line = $self->_get_line)) {
    chomp $read_line;
    $read_line =~ s/^\s+//;
    if ($continue) {
      $read_line =~ s/^\s+//;
      if ($read_line =~ /\\$/) {
        chop $read_line;
        $line .= $read_line;
      }
      else {
        $line .= $read_line;
        $self->_add_loaded_line($line);
        $continue = 0;
      }
      next;
    }
    next if $read_line eq '';
    next if $read_line =~ /^(#|!)/;
    if ($read_line =~ /\\$/) {
      chop $read_line;
      $line = $read_line;
      $continue = 1;
      next;
    }
    $self->_add_loaded_line($read_line);
  }
  if ($continue) {
    $self->_add_loaded_line($line);
  }
}


sub _load_file {
  my ($self, $file_name) = @_;

  my $handle = new FileHandle;
  $handle->open($file_name) || die "Can't open properties file '$file_name': $!\n";
  $self->{handle} = $handle;
  $self->_load;
  delete $self->{handle};
  $handle->close;
}


sub _add_loaded_line {
  my ($self, $line) = @_;

  my ($key, $value) = parse_line($line);
  $self->{property}->{$key} = $value;
}

sub parse_line {
  my $line = shift;

  if ($line =~ /^(.*?)((?<!\\)[\s=:].*)$/) {
    my $key = $1;
    my $rest = $2;
    $rest =~ s/^\s+//;
    $rest =~ s/^[:=]\s*//;
    return (unescape_key($key), unescape_value($rest));
  }
  else {
    return (unescape_key($line), undef);
  }
}

sub unescape_key {
  my $key = shift;

  $key =~ s/\\ / /g;
  $key =~ s/\\=/=/g;
  $key =~ s/\\:/:/g;

  return $key;
}

sub unescape_value {
  my $value = shift;

  $value =~ s/\\t/\t/g;
  $value =~ s/\\n/\n/g;
  $value =~ s/\\r/\r/g;
  $value =~ s/\\"/"/g;
  $value =~ s/\\'/'/g;
  $value =~ s/\\ / /g;
  $value =~ s/\\\\/\\/g;

  return $value;
}

package WiKicker::ReadOnlyProperties;
use 5.005;
use strict;
use Carp;

use vars qw(@ISA);
BEGIN {
  @ISA = qw(WiKicker::Properties);
}

sub set {
  croak "Can't set the value to Read Only Properties";
}

sub load {
  croak "Can't load to Read Only Properties";
}

sub load_file {
  croak "Can't load to ReadOnly Properties";
}

1;
