#
# WiKicker
#
#  Time.pm
#
#
# Copyright (C) 2003,2004 WATANABE Yoshimasa <naney@naney.org>
#
# This file is part of the WiKicker.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc., 59
# Temple Place, Suite 330, Boston, MA 02111-1307 USA
#
#
# $Id: Time.pm,v 1.15 2004/06/19 06:35:38 naney Exp $
#
#

package WiKicker::Time;

use 5.005;
use strict;
use Time::Zone;
use Time::Local;
use AutoLoader 'AUTOLOAD';

use vars qw(@wday_j_list @wday_e_list @mon_list @wday_class_list @month_days);

BEGIN {
  @wday_j_list = qw(日 月 火 水 木 金 土);
  @wday_e_list = qw(Sun Mon Tue Wed Thu Fri Sat);
  @mon_list = qw(dummy Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec);
  @wday_class_list = qw(Sunday Monday Tuesday Wednesday Thursday Friday Saturday);
  @month_days = (31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31);
}

sub _good_mday {
  my ($year, $month, $mday) = @_;

  my $range = $month_days[$month - 1];

  $range++ unless ($month != 2 || $year % 4 || !($year % 400));
  if (1 <= $mday && $mday <= $range) {
    return 1;
  }
  else {
    return 0;
  }
}

sub zonetime {
  my ($gmt_time, $zone) = @_;

  $zone =~ s/-\d+$//;
  unless($zone) {
    $zone = 'GMT';
  }

  my @result = gmtime($gmt_time + tz_offset($zone, $gmt_time));
  return @result;
}

sub _now_year {
  my $zone = shift;

  my ($now_sec, $now_min, $now_hour, $now_mday, $now_mon, $now_year, $now_wday, $now_yday, $now_isdst)
    = zonetime(time, $zone);

  return $now_year + 1900;
}

sub _wday_j_str {
  return $wday_j_list[shift];
}

sub _wday_e_str {
  return $wday_e_list[shift];
}

sub _wday_class_str {
  return $wday_class_list[shift];
}

sub _mon_str {
  return $mon_list[shift];
}

sub japanese_date_str {
  my ($gmt_time, $zone) = @_;

  my ($sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst)
    = zonetime($gmt_time, $zone);

  $year += 1900;
  $mon += 1;
  my $wday_j_str = _wday_j_str($wday);
  $mon = sprintf('%02d', $mon);
  $mday = sprintf('%02d', $mday);
  $hour = sprintf('%02d', $hour);
  $min = sprintf('%02d', $min);
  $sec = sprintf('%02d', $sec);

  if ($year == _now_year($zone)) {
    return "$mon月$mday日($wday_j_str) $hour:$min:$sec ($zone)";
  }
  else {
    return "$year年$mon月$mday日($wday_j_str) $hour:$min:$sec ($zone)";
  }
}

sub english_date_str {
  my ($gmt_time, $zone) = @_;

  my ($sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst)
    = zonetime($gmt_time, $zone);

  $year += 1900;
  $mon += 1;
  my $wday_e_str = _wday_e_str($wday);
  $mon = _mon_str($mon);
  $mday = sprintf('%02d', $mday);
  $hour = sprintf('%02d', $hour);
  $min = sprintf('%02d', $min);
  $sec = sprintf('%02d', $sec);

  if ($year == _now_year($zone)) {
    return qq($wday_e_str $mon $mday $hour:$min:$sec $zone);
  }
  else {
    return qq($wday_e_str $mon $mday $hour:$min:$sec $zone $year);
  }
}

sub japanese_date_html_str {
  my ($gmt_time, $zone) = @_;

  my ($sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst)
    = zonetime($gmt_time, $zone);

  $year += 1900;
  $mon += 1;
  my $wday_j_str = _wday_j_str($wday);
  my $wday_class = _wday_class_str($wday);
  $mon = sprintf('%02d', $mon);
  $mday = sprintf('%02d', $mday);
  $hour = sprintf('%02d', $hour);
  $min = sprintf('%02d', $min);
  $sec = sprintf('%02d', $sec);

  my $title = qq($year年$mon月$mday日($wday_j_str) $hour:$min:$sec ($zone));

  if ($year == _now_year($zone)) {
    return qq(<span title="$title">$mon月$mday日(<span class="$wday_class">$wday_j_str</span>)$hour:$min</span>);
  }
  else {
    return qq(<span title="$title">$year年$mon月$mday日(<span class="$wday_class">$wday_j_str</span>)$hour:$min</span>);
  }
}

sub english_date_html_str {
  my ($gmt_time, $zone) = @_;

  my ($sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst)
    = zonetime($gmt_time, $zone);

  $year += 1900;
  $mon += 1;
  my $wday_e_str = _wday_e_str($wday);
  my $wday_class = _wday_class_str($wday);
  $mon = _mon_str($mon);
  $mday = sprintf('%02d', $mday);
  $hour = sprintf('%02d', $hour);
  $min = sprintf('%02d', $min);
  $sec = sprintf('%02d', $sec);

  my $title = qq($wday_e_str $mon $mday $hour:$min:$sec $zone $year);

  if ($year == _now_year($zone)) {
    return qq(<span title="$title"><span class="$wday_class">$wday_e_str</span> $mon $mday $hour:$min</span>);
  }
  else {
    return qq(<span title="$title"><span class="$wday_class">$wday_e_str</span> $mon $mday $hour:$min $year</span>);
  }
}


sub _wday {
  my ($year, $month, $day) = @_;

  my $gmtime = timegm(0, 0, 0, $day, $month - 1, $year);

  my ($tmp_sec,$tmp_min,$tmp_hour,$tmp_mday,$tmp_mon,$tmp_year,$tmp_wday,$tmp_yday) = gmtime($gmtime);
  return $tmp_wday;
}

sub japanese_date_only_html_str {
  my ($year, $month, $day, $zone) = @_;

  unless (_good_mday($year, $month, $day)) {
    return "$year-$month-$day";
  }

  my $result = '';
  eval {
    my $wday = _wday($year, $month, $day);
    my $wday_j_str = _wday_j_str($wday);
    my $wday_class = _wday_class_str($wday);
    $month = sprintf('%02d', $month);
    $day = sprintf('%02d', $day);

    my $title = qq($year年$month月$day日($wday_j_str));

    if ($year == _now_year($zone)) {
      $result = qq(<span title="$title">$month月$day日(<span class="$wday_class">$wday_j_str</span>)</span>);
    }
    else {
      $result = qq(<span title="$title">$year年$month月$day日(<span class="$wday_class">$wday_j_str</span>)</span>);
    }
  };
  if ($@) {
    #warn $@;
    return "$year-$month-$day";
  }
  return $result;
}

sub english_date_only_html_str {
  my ($year, $month, $day, $zone) = @_;

  unless (_good_mday($year, $month, $day)) {
    return "$year-$month-$day";
  }

  my $result = '';
  eval {
    my $wday = _wday($year, $month, $day);
    my $wday_e_str = _wday_e_str($wday);
    my $wday_class = _wday_class_str($wday);
    my $mon = _mon_str($month);
    $day = sprintf('%02d', $day);

    my $title = qq($wday_e_str $mon $day $year);

    if ($year == _now_year($zone)) {
      $result = qq(<span title="$title"><span class="$wday_class">$wday_e_str</span> $mon $day</span>);
    }
    else {
      $result = qq(<span title="$title"><span class="$wday_class">$wday_e_str</span> $mon $day $year</span>);
    }
  };
  if ($@) {
    #warn $@;
    return "$year-$month-$day";
  }
  return $result;
}

sub gmtime_to_RFC1123 {
  my $time = shift;

  my @WDAY = qw/Sun Mon Tue Wed Thu Fri Sat/;
  my @MON = qw/Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec/;

  my($sec,$min,$hour,$mday,$mon,$year,$wday) = gmtime($time);
  $year += 1900;
  return sprintf('%s, %02d %s %04d %02d:%02d:%02d GMT',
                 $WDAY[$wday],$mday,$MON[$mon],$year,$hour,$min,$sec);
}

sub RFC1123_to_gmtime {
  my $time_str = shift;

  unless ($time_str =~ /^(.+), (\d\d) (.+) (\d\d\d\d) (\d\d):(\d\d):(\d\d) GMT$/) {
    return 0;
  }
  my ($mday, $mon_str, $year, $hour, $min, $sec) = ($2, $3, $4, $5, $6, $7);

  my %MON = (Jan => 0,
             Feb => 1,
             Mar => 2,
             Apr => 3,
             May => 4,
             Jun => 5,
             Jul => 6,
             Aug => 7,
             Sep => 8,
             Oct => 9,
             Nov => 10,
             Dec => 11);
  unless (exists $MON{$mon_str}) {
    return 0;
  }
  my $mon = $MON{$mon_str};

  return timegm($sec, $min, $hour, $mday, $mon, $year);
}

sub gmtime_to_W3CDTF_UTC;
sub W3CDTF_to_gmtime;
sub _W3CDTF_to_gmtime;
sub japanese_date_only_str;
sub english_date_only_str;
sub day_in_ten_day;
sub previous_day_str;

1;

__END__
sub gmtime_to_W3CDTF_UTC {
  my $datetime = shift;

    my ($sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst)
    = gmtime($datetime);
    $year += 1900;
    $mon += 1;
    return sprintf("%d-%02d-%02dT%02d:%02d:%02dZ",
                   $year, $mon, $mday, $hour, $min, $sec);
}

sub W3CDTF_to_gmtime {
  my $datetime = shift;
  my $result = '';
  eval {
    $result = _W3CDTF_to_gmtime($datetime);
  };
  if ($@) {
    #warn $@;
    return undef;
    #return -1;
  }
  return $result;
}

sub _W3CDTF_to_gmtime {
  my $datetime = shift;

  unless ($datetime =~ /^(\d\d\d\d)(.*)$/) {
    return undef;
  }
  my $year = $1;
  my $rest = $2;
  if (length($rest) == 0) {
    return timegm(0, 0, 0, 1, 0, $year);
  }

  unless ($rest =~ /^-(\d\d)(.*)$/) {
    return undef;
  }
  my $mon = $1;
  $rest = $2;
  if (length($rest) == 0) {
    return timegm(0, 0, 0, 1, $mon - 1, $year);
  }

  unless ($rest =~ /^-(\d\d)(.*)$/) {
    return undef;
  }
  my $mday = $1;
  $rest = $2;

  unless (_good_mday($year, $mon, $mday)) {
    return undef;
  }

  if (length($rest) == 0) {
    return timegm(0, 0, 0, $mday, $mon - 1, $year);
  }

  unless ($rest =~ /^T(\d\d):(\d\d)(:(\d\d)(.\d+)?)?(Z|([+-]\d\d:\d\d))$/) {
    return undef;
  }
  my ($hour, $min, $secstr, $sec, $zone) = ($1, $2, $3, $4, $6);
  unless ($secstr) {
    $sec = 0;
  }
  my $zone_value;
  if ($zone eq 'Z') {
    $zone_value = 0;
  }
  elsif ($zone =~ /^\+(\d\d):(\d\d)$/) {
    $zone_value = ($1 * 60 + $2) * 60;
  }
  elsif ($zone =~ /^\-(\d\d):(\d\d)$/) {
    $zone_value = ($1 * 60 + $2) * -60;
  }
  else {
    die;
  }

  return timegm($sec, $min, $hour, $mday, $mon - 1, $year) - $zone_value;
}

sub japanese_date_only_str {
  my ($year, $month, $day, $zone) = @_;

  my $result = '';
  unless (_good_mday($year, $month, $day)) {
    return "$year-$month-$day";
  }
  eval {
    my $wday = _wday($year, $month, $day);
    my $wday_j_str = _wday_j_str($wday);
    $month = sprintf('%02d', $month);
    $day = sprintf('%02d', $day);

    if ($year == _now_year($zone)) {
      $result = "$month月$day日($wday_j_str)";
    }
    else {
      $result = "$year年$month月$day日($wday_j_str)";
    }
  };
  if ($@) {
    #warn $@;
    return "$year-$month-$day";
  }
  return $result;
}

sub english_date_only_str {
  my ($year, $month, $day, $zone) = @_;

  my $result = '';
  unless (_good_mday($year, $month, $day)) {
    return "$year-$month-$day";
  }
  eval {
    my $wday = _wday($year, $month, $day);
    my $wday_e_str = _wday_e_str($wday);
    my $month_string = _mon_str($month);
    $day = sprintf('%02d', $day);

    if ($year == _now_year($zone)) {
      $result = "$wday_e_str $month_string $day";
    }
    else {
      $result = "$wday_e_str $month_string $day $year";
    }
  };
  if ($@) {
    #warn $@;
    return "$year-$month-$day";
  }
  return $result;
}

sub day_in_ten_day {
  my ($day, $ten_day) = @_;

  if ($ten_day eq 'a') {
    return (1 <= $day && $day <= 10);
  }
  elsif ($ten_day eq 'b') {
    return (11 <= $day && $day <= 20);
  }
  elsif ($ten_day eq 'c') {
    return (21 <= $day && $day <= 31);
  }
}

sub previous_day_str {
  my $date_str = shift;

  $date_str =~ /^(\d\d\d\d)-(\d\d)-(\d\d)$/;
  my ($y, $m, $d) = ($1, $2, $3);

  if ($d > 1) {
    $d--;
    return sprintf("%4d-%02d-%02d", $y, $m, $d);
  }
  else {
    $m--;
    if ($m == 0) {
      $y--;
      $m = 12;
    }
    $d = $month_days[$m - 1];
    $d++ unless ($m != 2 || $y % 4 || !($y % 400));
  }
  return sprintf("%4d-%02d-%02d", $y, $m, $d);
}
