#
# WiKicker
#
#  Resolver.pm
#
#
# Copyright (C) 2003,2004 WATANABE Yoshimasa <naney@naney.org>
#
# This file is part of the WiKicker.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc., 59
# Temple Place, Suite 330, Boston, MA 02111-1307 USA
#
#
# $Id: Resolver.pm,v 1.15 2004/02/10 13:19:42 naney Exp $
#
#

package WiKicker::WRI::Resolver;

use strict;
use WiKicker::URI;
use WiKicker::WRI::RWRN;
use WiKicker::UTF8;

sub new {
  my $class = shift;

  my $self = bless { interwiki => [],
                   }, $class;

  return $self;
}

sub db {
  my ($self, $value) = @_;

  if (defined $value) {
    $self->{db} = $value;
  }
  return $self->{db};
}

sub amazon_id {
  my ($self, $value) = @_;

  if (defined $value) {
    $self->{amazon_id} = $value;
  }
  return $self->{amazon_id};
}

sub add_interwiki {
  my ($self, @wri_list) = @_;

  push(@{$self->{interwiki}}, @wri_list);
}

sub get_uwrn_link_url {
  my ($self, $wri) = @_;

  my $definition = $self->find_interwiki_definition($wri->scheme);
  return '' unless $definition;

  return $self->apply_interwiki_definition($definition, $wri);
}

sub get_interwiki_link_url {
  my ($self, $scheme, $name) = @_;

  my $definition = $self->find_interwiki_definition($scheme);
  return '' unless $definition;

  return $self->apply_interwiki_definition_to_text($definition, $name);
}

sub find_interwiki_definition {
  my ($self, $qualifier) = @_;

  for my $definition (@{$self->{interwiki}}) {
    my $definition_qualifier = $definition->nid;
    if ($qualifier eq $definition_qualifier) {
      return $definition;
    }
  }
  if (defined $self->db &&
      $self->db->exists_cache('interwiki', $qualifier)) {
    my $definition = $self->db->cache('interwiki', $qualifier);
    my $wri = new WiKicker::WRI::RWRN("interwiki:$qualifier" . $definition,
                                      'interwiki', "$qualifier:$definition");
    $self->add_interwiki($wri);
    return $wri;
  }
  else {
    return undef;
  }
}

sub apply_interwiki_definition {
  my ($self, $definition, $uwrn) = @_;

  my $rest = $definition->rest;
  $rest =~ /^([^:]+):(.*)$/;
  my ($qualifier, $body) = ($1, $2);
  my $name = $uwrn->rest;

  $body =~ s/\$\{(.+?)\}/$self->_expand($1, $name)/eg;

  return $body;
}

sub apply_interwiki_definition_to_text {
  my ($self, $definition, $name) = @_;

  my $rest = $definition->rest;
  $rest =~ /^([^:]+):(.*)$/;
  my ($qualifier, $body) = ($1, $2);

  $body =~ s/\$\{(.+?)\}/$self->_expand($1, $name)/eg;

  return $body;
}

my %expand_table =
  ('name'      => sub { return shift; },
   'name.uri'  => sub { return WiKicker::URI::escape_URI(shift); },
   'name.path' =>
            sub { return WiKicker::URI::page_name_to_path_segments(shift); },
   'name.urlencoded' =>
            sub { return WiKicker::URI::encode_x_www_form_urlencoded(shift); },
   'name.PukiWiki1.3' =>
            sub { my $name = shift;
                  if ($name =~ /^([A-Z][a-z]+){2,}$/) {
                    return $name;
                  }
                  $name = WiKicker::UTF8::encode('euc', $name);
                  return WiKicker::URI::escape_URI("[[$name]]");
                },
   'name.euc' =>
            sub { my $name = shift;
                  $name = WiKicker::UTF8::encode('euc', $name);
                  return $name;
                },
   'name.EUC-JP' =>
            sub { my $name = shift;
                  $name = WiKicker::UTF8::encode('euc', $name);
                  return $name;
                },
   'name.euc.uri' =>
            sub { my $name = shift;
                  $name = WiKicker::UTF8::encode('euc', $name);
                  return WiKicker::URI::escape_URI($name);
                },
   'name.EUC-JP.uri' =>
            sub { my $name = shift;
                  $name = WiKicker::UTF8::encode('euc', $name);
                  return WiKicker::URI::escape_URI($name);
                },
   'name.euc.urlencoded' =>
            sub { my $name = shift;
                  $name = WiKicker::UTF8::encode('euc', $name);
                  return WiKicker::URI::encode_x_www_form_urlencoded($name);
                },
   'name.EUC-JP.urlencoded' =>
            sub { my $name = shift;
                  $name = WiKicker::UTF8::encode('euc', $name);
                  return WiKicker::URI::encode_x_www_form_urlencoded($name);
                },
   'name.Shift_JIS' =>
            sub { my $name = shift;
                  $name = WiKicker::UTF8::encode('sjis', $name);
                  return $name;
                },
   'name.Shift_JIS.uri' =>
            sub { my $name = shift;
                  $name = WiKicker::UTF8::encode('sjis', $name);
                  return WiKicker::URI::escape_URI($name);
                },
   'name.Shift_JIS.urlencoded' =>
            sub { my $name = shift;
                  $name = WiKicker::UTF8::encode('sjis', $name);
                  return WiKicker::URI::encode_x_www_form_urlencoded($name);
                },
);   

sub _expand {
  my ($self, $variable, $name) = @_;

  if (exists $expand_table{$variable}) {
    return &{$expand_table{$variable}}($name);
  }
  return $name;
}

sub get_rfc_url {
  my ($self, $number) = @_;

  return "http://www.ietf.org/rfc/rfc$number.txt";
}

sub get_asin_url {
  my ($self, $asin) = @_;

  unless ($self->amazon_id) {
    return '';
  }

  return 'http://www.amazon.co.jp/exec/obidos/ASIN/'
    . $asin . '/' . $self->amazon_id . '/ref=nosim';
}  


sub get_asin_image_url {
  my ($self, $asin) = @_;

  unless ($self->amazon_id) {
    return '';
  }

  my $n;
  if ($asin =~ /^[01]/) {
    $n = '01';
  }
  elsif ($asin eq 'B0000645I7' || $asin eq 'B00008W2OO' || $asin eq 'B00004R7VU') {
    $n = '01';
  }
  else {
    $n = '09';
  }
  $asin = WiKicker::URI::escape_URI($asin);
  return qq(http://images-jp.amazon.com/images/P/$asin.$n.MZZZZZZZ.jpg);
}

sub get_isbn_image_url {
  my ($self, $isbn) = @_;

  unless ($self->amazon_id) {
    return '';
  }
  
  my $n;
  if ($isbn =~ /^0[0567]/) {
    $n = '01';
  }
  elsif ($isbn =~ /^1[59]/) {
    $n = '01';
  }
  else {
    $n = '09';
  }
  $isbn = WiKicker::URI::escape_URI($isbn);
  return qq(http://images-jp.amazon.com/images/P/$isbn.$n.MZZZZZZZ.jpg);
}


1;
