#
# WiKicker
#
#  UriTitleDatabase.pm
#
#
# Copyright (C) 2003 WATANABE Yoshimasa <naney@naney.org>
#
# This file is part of the WiKicker.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc., 59
# Temple Place, Suite 330, Boston, MA 02111-1307 USA
#
#
# $Id: UriTitleDatabase.pm,v 1.2 2004/01/29 14:37:35 naney Exp $
#
#

package WiKicker::WWW::UriTitleDatabase;

use strict;
use Storable;
use MLDBM;
use MLDBM::Sync::SDBM_File;
use WiKicker::Tool::GetUriTitle;
use WiKicker::URI;
use Fcntl;
use WiKicker::WWW::Robot::RobotUserAgent;
use WiKicker::HTTP::Referer;
use LWP;
use WiKicker::UTF8;

sub new {
  my ($class, %param) = @_;
  my $robot_rules_db_file_name = $param{robot_rules_db_file_name};
  my $uri_title_db_file_name = $param{uri_title_db_file_name};
  my $max_get = $param{max_get} || 0;
  my $max_time = $param{max_time} || 0;
  my $agent = $param{agent} || "libwww-perl/$LWP::VERSION";

  my $self = bless {}, $class;
  $self->{robot_rules_db_file_name} = $robot_rules_db_file_name;
  $self->{uri_title_db_file_name} = $uri_title_db_file_name;
  $self->{user_agent} = WiKicker::WWW::Robot::RobotUserAgent->new($agent,
                                       $param{from},
                                       $robot_rules_db_file_name);
  $self->{get_count} = 0;
  $self->{max_get} = $max_get;
  $self->{max_time} = $max_time;
  $self->{start_time} = time;

  return $self;
}

sub get_title {
  my ($self, $uri) = @_;

  $uri =~ s/#.*$//;

  return undef unless $uri =~ m|^http://|;

  my $title = $self->_get_title_from_uri($uri);
  if (defined $title) {
    return $title;
  }

  my $last_modified = undef;
  my $old_title = undef;
  my ($title, $old_title, $last_modified, $skip)
    = $self->_get_title_from_db($uri);
  if (defined $title) {
    return $title;
  }
  if (defined $skip) {
    #warn "SKIP(may not be change): $uri";
    return $title;
  }
  
  if ($self->{max_get} && $self->{get_count} >= $self->{max_get}) {
    #warn "TIME OVER: $uri";
    $self->_user_agent->add($uri);
    return undef;
  }
  if ($self->{max_time} && $self->{start_time} + $self->{max_time} <= time) {
    #warn "TIME OVER: $uri";
    $self->_user_agent->add($uri);
    return undef;
  }
  $self->{get_count}++;
  return $self->_get_title_from_server($uri, $old_title, $last_modified);
}

sub print_queue {
  my ($self, $handle) = @_;

  $self->_user_agent->print_queue($handle);
}

sub _user_agent {
  my $self = shift;

  return $self->{user_agent};
}

sub _db_hash {
  my $self = shift;

  if (exists $self->{db_hash}) {
    return $self->{db_hash};
  }

  $MLDBM::UseDB = 'MLDBM::Sync::SDBM_File';
  $MLDBM::Serializer = 'Storable';
  tie(%{$self->{db_hash}},
      'MLDBM',
      $self->{uri_title_db_file_name},
      O_CREAT|O_RDWR,
      0640) || die $!;

  return $self->{db_hash};
}

sub _debug_print {
  my $self = shift;

  my $db_hash = $self->_db_hash;
  foreach (keys %$db_hash) {
    next if $_ eq 'queue';
    my $tmp = $db_hash->{$_};
    my $title = $tmp->{title};
    $title = WiKicker::UTF8::encode('euc', $title);
    #warn $_, ' ', $title, ' ', $tmp->{code}, "\n";
  }
}

sub _debug_print_all_title {
  my $self = shift;

  my $db_hash = $self->_db_hash;
  foreach (keys %$db_hash) {
    next if $_ eq 'queue';
    my $tmp = $db_hash->{$_};
    next unless $tmp->{title};
    my $title = $tmp->{title};
    $title = WiKicker::UTF8::encode('euc', $title);
    print "$_ $title\n";
  }
}

sub _get_title_from_uri {
  my ($self, $uri) = @_;
  
  my $referer = WiKicker::HTTP::Referer->new($uri);
  my ($engine, $query) = $referer->get_search_query;
  if (defined $engine) {
    return "$engine:$query";
  }

  return undef;
}

sub next_check_time {
  my ($self, $uri) = @_;

  my $db_hash = $self->_db_hash;
  #unless (exists $db_hash->{$uri}) {
  unless (defined $db_hash->{$uri}) {
    return 0;
  }
  my $info = $db_hash->{$uri};
  return $self->_next_check_time($info);
}

sub _next_check_time {
  my ($self, $info) = @_;

  if ($info->{tk} < 3600 * 24) {
    return $info->{tc} + 3600 * 24;
  }
  if ($info->{tk} * 2 > 3600 * 24 * 30) {
    return $info->{tc} + $info->{tk} + 3600 * 24 * 30;
  }
  return$info->{tc} + $info->{tk} * 2;
}

sub _get_title_from_db {
  my ($self, $uri) = @_;

  my $db_hash = $self->_db_hash;
  # exists?
  #unless (exists $db_hash->{$uri}) {
  unless (defined $db_hash->{$uri}) {
    return (undef, undef, undef);
  }

  my $info = $db_hash->{$uri};
  # fresh?
  if (time <= $info->{fu}) {
    $self->_touch_db($uri);
    return ($info->{t}, $info->{t}, $info->{lm});
  }
  else {
    # sorosoro title change

    if ($self->_next_check_time($info) < time) {
      return (undef, $info->{t}, $info->{lm});
    }
    else {
      # may be not changed
      return ($info->{t}, $info->{t}, $info->{lm}, 1);
    }

    # db is old
    return (undef, $info->{t}, $info->{lm});
  }
}

sub get_uri_info {
  my ($self, $uri) = @_;

  $uri =~ s/#.*$//;
  
  my $db_hash = $self->_db_hash;
  # exists?
  #unless (exists $db_hash->{$uri}) {
  unless (defined $db_hash->{$uri}) {
    return undef;
  }

  my $info = $db_hash->{$uri};
  return { uri => $uri,
           last_access => $info->{la},
           title => $info->{t},
           code => $info->{c},
           last_modified => $info->{lm},
           fresh_until => $info->{fu},
           title_changed => $info->{tc},
           title_kept => $info->{tk},
           get_count => $info->{g} || 1,
         };
}

sub _touch_db {
  my ($self, $uri) = @_;

  my $db_hash = $self->_db_hash;
  my $tmp = $db_hash->{$uri};
  $tmp->{la} = time; # la = last access
  $db_hash->{$uri} = $tmp;
}

sub _set_uri_information {
  my ($self, $uri, $title, $code, $last_modified, $fresh_until, $no_count_up) = @_;

  my $db_hash = $self->_db_hash;

  my $now = time;

  my $title_changed;
  my $title_kept;
  my $get_count = 0;
  #if (exists $db_hash->{$uri}) {
  if (defined $db_hash->{$uri}) {
    my $info = $db_hash->{$uri};
    $get_count = $info->{g} || 0;
    if (defined $info->{t}) {
      if (defined $title && $info->{t} eq $title) {
        $title_changed = $info->{tc};
        $title_kept = $now - $title_changed;
      }
      else {
        $title_changed = $now;
        $title_kept = 0;
      }
    }
    else {
      if (defined $title) {
        $title_changed = $now;
        $title_kept = 0;
      }
      else {
        $title_changed = $info->{tc};
        $title_kept = $now - $title_changed;
      }
    }
  }
  else {
    $title_changed = $now;
    $title_kept = 0;
  }
  unless (defined $no_count_up && $no_count_up) {
    $get_count++;
  }
  
  $db_hash->{$uri} = {la => $now,
                      t => $title,
                      tc => $title_changed,
                      tk => $title_kept,
                      c => $code,
                      lm =>$last_modified,
                      fu => $fresh_until,
                      g => $get_count,
                     };
}

sub _get_title_from_server {
  my ($self, $uri, $old_title, $last_modified) = @_;

  my $response = $self->_user_agent->request($uri, $last_modified);

  unless (defined $response) {
    #warn "RETRY AFTER: $uri\n";
    return undef;
  }

  my $title = undef;
  if ($response->is_success) {
    $title = $self->_get_title_from_response($response, $uri);
    #warn "GET!: $uri\n";
    $self->_set_uri_information($uri, $title, $response->code,
                                $response->last_modified,
                                $response->fresh_until);
  }
  elsif ($response->code == HTTP::Status::RC_NOT_MODIFIED) {
    $self->_set_uri_information($uri, $old_title, $response->code,
                                $response->last_modified,
                                $response->fresh_until);
  }
  else {
    my $no_count_up = 0;
    if ($response->code == HTTP::Status::RC_FORBIDDEN
        && $response->message =~ /robots.txt/) {
      $no_count_up = 1;
      #warn "DISALLOW: $uri\n";
    }
    else {
      #warn "ERROR: $uri\n";
    }
    $self->_set_uri_information($uri, undef, $response->code,
                                $response->last_modified,
                                $response->fresh_until,
                                $no_count_up
                               );
  }
  return $title;
}

sub _get_title_from_response {
  my ($self, $response, $uri) = @_;
  
  my $content_type = $response->content_type;
  if ($content_type && $content_type ne 'text/html') {
    return '';
  }
  unless ($content_type) {
    unless ($uri =~ /\.html?$/) {
      return '';
    }
  }
  my $content = $response->content;
  return WiKicker::Tool::GetUriTitle->new->get_title_from_html($content);
}

1;
