#
# WiKicker
#
#  Request.pm
#
#
# Copyright (C) 2003,2004 WATANABE Yoshimasa <naney@naney.org>
#
# This file is part of the WiKicker.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc., 59
# Temple Place, Suite 330, Boston, MA 02111-1307 USA
#
#
# $Id: Request.pm,v 1.5 2004/08/03 12:55:52 naney Exp $
#
#

package WiKicker::WikiCGI::Request;

use 5.005;
use strict;
use Carp;
use WiKicker::CGI::User;
use WiKicker::CGI::Cookie;
use WiKicker::PageName;
use WiKicker::URI;

use WiKicker::CGI::Request;
use AutoLoader 'AUTOLOAD';

use vars qw(@ISA);

BEGIN {
  @ISA = qw(WiKicker::CGI::Request);
}


my %normal_page_action =
  (
   read => 1,
   edit => 1,
   search => 1,
   rlog => 1,
   delete => 1,
   diff => 1,
  );


sub new {
  my ($class, %options) = @_;

  my $self =
    bless
      {
       name => '',
       action => 'read',
       format => 'html'
      }, $class;

  if (exists $options{environment}) {
    $self->environment($options{environment});
  }
  else {
    confess 'no environment';
  }

  if (exists $options{query}) {
    $self->_set($options{query});
  }

  my $cookies = fetch CGI::Cookie;
  $self->cookies($cookies);

  $self->_initialize_user;

  return $self;
}

sub name {
  my ($self, $value) = @_;

  if (defined $value) {
    $value = WiKicker::PageName::normalize($value);
    $self->{name} = $value;
  }
  return $self->{name};
}


sub _set {
  my ($self, $query) = @_;

  $self->_set_query($query);

  if ($query->path_info) {
    $self->_set_with_path_info($query);
  }
  else {
    $self->_set_without_path_info($query);
  }
}

sub _set_with_path_info {
  my ($self, $query) = @_;

  $self->action('read'); # default;

  my $path_info = $query->path_info;
  $path_info =~ s|^/||;
  if ($path_info =~ /(.*)\.([^.]+)$/) {
    my $name = $1;
    my $format = $2;
    $self->name($name);
#    if ($format eq 'TrackBack') {
#      $self->format('xml');
#      $self->action('TrackBack');
#    }
#    else {
      $self->format($format);
#    }
  }
  else {
    $self->name($path_info);
    $self->format('html');
  }

  if ($query->keywords) {
    $self->action(join('', $query->keywords));
    return;
  }

  if ($query->param('action')) {
    $self->action($query->param('action'));
  }
}

sub _set_without_path_info {
  my ($self, $query) = @_;

  $self->name('FrontPage'); #default

  if ($query->keywords) {
    $self->name(join('', $query->keywords));
    return;
  }

  if ($query->param('action')) {
    $self->action($query->param('action'));
    $self->name($query->param('name'));
  }
  else {
    my @control_names = $query->param;
    my $action = $self->_get_action_from_control_names(@control_names);
    if ($action) {
      $self->action($action);
      $self->name($query->param($action));
    }
    else {
      if ($query->param('name')) {
        $self->action('read');
        $self->name($query->param('name'));
      }
    }
  }

  if ($query->param('format')) {
    $self->format($query->param('format'));
  }
  else {
    $self->format('html');
  }
}

sub _get_action_from_control_names {
  my ($self, @control_names) = @_;

  my $result = '';
  for my $control_name (@control_names) {
    if (exists $normal_page_action{$control_name}) {
      if ($result) {
        return ''; # multiple action
      }
      else {
        $result = $control_name;
      }
    }
  }
  return $result;
}

sub format {
  my ($self, $value) = @_;

  if (defined $value) {
    $self->{format} = $value;
  }
  return $self->{format};
}

sub is_default_view_user {
  my $self = shift;

  if (exists $self->{_is_default_view_user}) {
    return $self->{_is_default_view_user};
  }

  my $user = $self->user;
  my $default_timezone = WiKicker::User::DEFAULT_USER_TIMEZONE;
  if ($self->configuration) {
    $default_timezone = $self->configuration->properties->get('user.timezone');
  }
  if ($user->name eq '' &&
      $user->editor_width == WiKicker::User::DEFAULT_USER_EDITOR_WIDTH &&
      $user->editor_height == WiKicker::User::DEFAULT_USER_EDITOR_HEIGHT &&
      $user->timezone eq $default_timezone) {
    return $self->{_is_default_view_user} = 1;
  }
  return $self->{_is_default_view_user} = 0;
}

sub is_self_uri;

1;

__END__

sub is_self_uri {
  my ($self, $uri, $self_uri) = @_;

  my $host_port;
  my $script_name;
  if (defined $self_uri) {
    my $self_uri_obj = URI->new($self_uri);
    $host_port = $self_uri_obj->host_port;
    $script_name = $self_uri_obj->path;
  }
  else {
    $host_port = $ENV{SERVER_NAME} . ':' . $ENV{SERVER_PORT};
    $script_name = $ENV{SCRIPT_NAME};
  }

  $host_port = lc($host_port);
  my $target = URI->new($uri);
  if (defined $target->scheme) {
    unless ($target->scheme eq 'http' || $target->scheme eq 'https') {
      return undef;
    }
  }
  else {
    $target->scheme('http'); # default
  }

  if (defined $target->host_port) {
    unless (lc($target->host_port) eq $host_port) {
      return undef;
    }
  }

  unless (defined $target->path) {
    return undef;
  }

  my $path_info = '';
  if ($target->path =~ m|^\Q$script_name\E(.*)$|) {
    $path_info = $1 || '';
  }
  else {
    return undef;
  }

  if ($path_info ne '') {
    if (substr($path_info, 0, 1) ne '/') {
      return undef;
    }
    $path_info = substr($path_info, 1);

    if ($path_info =~ /(.*)\.([^.]+)$/) {
      my $name = WiKicker::URI::unescape_URI($1);
      my $format = $2;

      return $name;
    }
    else {
      return WiKicker::URI::unescape_URI($path_info);
    }
  }

  if (defined $target->query_keywords) {
    return join('', $target->keywords);
  }

  my %query = $target->query_form;
  if (exists $query{action}) {
    return WiKicker::URI::unescape_URI($query{name});
  }
  else {
    my @control_names = keys %query;
    my $action = $self->_get_action_from_control_names(@control_names);
    if ($action) {
      return $query{action};
    }
    else {
      if (exists $query{name}) {
        return WiKicker::URI::unescape_URI($query{name});
      }
    }
  }

  return '';
}
