#
# WiKicker
#
#  DocumentParser.pm
#
#
# Copyright (C) 2003,2004 WATANABE Yoshimasa <naney@naney.org>
#
# This file is part of the WiKicker.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc., 59
# Temple Place, Suite 330, Boston, MA 02111-1307 USA
#
#
# $Id: DocumentParser.pm,v 1.8 2004/08/03 12:55:52 naney Exp $
#
#

package WiKicker::WikiPage::DocumentParser;

use strict;
use Carp;
use WiKicker::WikiPage::DocumentNode;
use WiKicker::WikiPage::ParagraphNode;
use WiKicker::WikiPage::PreNode;
use WiKicker::WikiPage::EmptyLinesNode;
use WiKicker::WikiPage::HeadingNode;
use WiKicker::WikiPage::ListNode;
use WiKicker::WikiPage::ListItemNode;
use WiKicker::WikiPage::HrNode;
use WiKicker::WikiPage::InlineParser;
use WiKicker::WikiPage::QuoteNode;
use WiKicker::WikiPage::QuoteLineNode;
use WiKicker::Loader;
use AutoLoader 'AUTOLOAD';

sub new {
  my ($class, %param) = @_;

  croak 'no wri_factory' unless defined $param{wri_factory};

  my $self = bless { current_path => [],
                     wri_factory => $param{wri_factory},
                   }, $class;

  return $self;
}

sub document_node {
  my ($self, $value) = @_;

  if (defined $value) {
    $self->{document_node} = $value;
  }
  return $self->{document_node};
}

sub current_node {
  return shift->{current_path}->[-1];
}

sub push_node {
  my ($self, $node) = @_;

  $self->current_node->add($node);
  push(@{$self->{current_path}}, $node);
}

sub pop_node {
  my $self = shift;

  return pop(@{$self->{current_path}});
}

sub clear_node_stack {
  my $self = shift;

  #while (@{$self->{current_path}} > 1) {
  #  pop(@{$self->{current_path}});
  #}
  splice(@{$self->{current_path}}, 1);
}

sub parse {
  my ($self, $text) = @_;

  $text =~ s/\x0d\x0a/\x0a/g;
  $self->document_node(WiKicker::WikiPage::DocumentNode->new);
  $self->{current_path} = [$self->document_node];
  $self->_parse($text);
  return $self->document_node;
}

sub get_token {
  my ($self, $line) = @_;

  if ($line =~ /^\s*$/) {
    return ('empty', $line);
  }
  if ($line =~ /^(  \*{1,3}
                  | ----
                  | -{1,3}
                  | \#{1,3}
                  | >{1,3}
                  | :{1,3}
                  | ,
                  | \|
                  | \s)/ogx) {
    my $type = $1;
    return ($type, substr($line, length($type)));
  }
  return ('paragraph', $line);
}

sub _parse {
  my ($self, $text) = @_;

  my @lines = split(/(\n)/, $text);
  while (@lines) {
    my $line = shift @lines;
    $line .= shift @lines if @lines;
    my ($type, $data) = $self->get_token($line);

    $self->do_empty($data) && next if $type eq 'empty';
    $self->do_paragraph($data) && next if $type eq 'paragraph';
    $self->do_uli(1, $data) && next if $type eq '-';
    $self->do_uli(2, $data) && next if $type eq '--';
    $self->do_uli(3, $data) && next if $type eq '---';
    $self->do_hr($data) && next if $type eq '----';
    $self->do_heading(1, $data) && next if $type eq '*';
    $self->do_heading(2, $data) && next if $type eq '**';
    $self->do_heading(3, $data) && next if $type eq '***';
    $self->do_oli(1, $data) && next if $type eq '#';
    $self->do_oli(2, $data) && next if $type eq '##';
    $self->do_oli(3, $data) && next if $type eq '###';
    $self->do_quote(1, $data) && next if $type eq '>';
    $self->do_quote(2, $data) && next if $type eq '>>';
    $self->do_quote(3, $data) && next if $type eq '>>>';
    $self->do_dl(1, $data) && next if $type eq ':';
    $self->do_dl(2, $data) && next if $type eq '::';
    $self->do_dl(3, $data) && next if $type eq ':::';
    $self->do_table($data, $type) && next if $type eq '|' || $type eq ',';
    $self->do_pre($data) && next if $type eq ' ';
    die "Internal error: unknown type '$type'";
  }
}

sub _parse_inline {
  my ($self, $text) = @_;

  return WiKicker::WikiPage::InlineParser
    ->new(wri_factory => $self->{wri_factory})->parse($text);
}

sub _trim {
  my $str = shift;

  $str =~ m/^(\s*)(.*?)(\s*)$/;
  return ($1, $2, $3);
}

sub do_paragraph {
  my ($self, $data) = @_;

  if (ref $self->current_node ne 'WiKicker::WikiPage::ParagraphNode') {
    $self->clear_node_stack;
    $self->push_node(WiKicker::WikiPage::ParagraphNode->new);
  }

  $self->current_node->add($self->_parse_inline($data));
}

sub do_empty {
  my ($self, $data) = @_;

  if (ref $self->current_node ne 'WiKicker::WikiPage::EmptyLinesNode') {
    $self->clear_node_stack;
    $self->push_node(WiKicker::WikiPage::EmptyLinesNode->new);
  }

  $self->current_node->add($data);
}


sub do_heading {
  my ($self, $level, $data) = @_;

  my $heading_node = WiKicker::WikiPage::HeadingNode->new;
  $heading_node->level($level);
  $heading_node->add($self->_parse_inline($data));
  $self->clear_node_stack;
  $self->push_node($heading_node);
}

sub do_hr {
  my ($self, $data) = @_;

  my $hr_node = WiKicker::WikiPage::HrNode->new('----'. $data);
  $self->clear_node_stack;
  $self->push_node($hr_node);
}

sub do_uli {
  my ($self, $level, $data) = @_;

  $self->set_list_node_to_stack_top($level, 0);
  $self->current_node->add($self->_make_list_item_node($level, 0, $data));
}

sub do_oli {
  my ($self, $level, $data) = @_;

  $self->set_list_node_to_stack_top($level, 1);
  $self->current_node->add($self->_make_list_item_node($level, 1, $data));
}

sub set_list_node_to_stack_top {
  my ($self, $level, $ordered) = @_;

  while (1) {
    if (ref $self->current_node eq 'WiKicker::WikiPage::DocumentNode') {
      my $list_node = WiKicker::WikiPage::ListNode->new;
      $list_node->ordered($ordered);
      $list_node->level(1);
      $self->push_node($list_node);
      next;
    }
    elsif (ref $self->current_node eq 'WiKicker::WikiPage::ListNode') {
      my $current_level = $self->current_node->level;
      if ($current_level == $level) {
        if ($self->current_node->ordered == $ordered) {
          return;
        }
        else {
          $self->pop_node;
          next;
        }
      }
      elsif ($current_level > $level) {
        $self->pop_node;
        next;
      }
      else {
        my @nodes = $self->current_node->nodes;
        if (@nodes > 0 && ref $nodes[-1] eq 'WiKicker::WikiPage::ListItemNode') {
          my $list_node = WiKicker::WikiPage::ListNode->new;
          $list_node->ordered($ordered);
          $list_node->level($current_level + 1);
          $nodes[-1]->add($list_node);
          push(@{$self->{current_path}}, $list_node);
          next;
        }
        else {
          my $list_item_node = WiKicker::WikiPage::ListItemNode->new;
          $list_item_node->level($current_level);
          $list_item_node->ordered($self->current_node->ordered);
          $list_item_node->dummy(1);
          $self->current_node->add($list_item_node);

          my $list_node = WiKicker::WikiPage::ListNode->new;
          $list_node->ordered($ordered);
          $list_node->level($current_level + 1);

          $list_item_node->add($list_node);
          push(@{$self->{current_path}}, $list_node);
          next;
        }
      }
    }
    $self->clear_node_stack;
  }
}

sub _make_list_item_node {
  my ($self, $level, $ordered, $data) = @_;
  my $list_item_node = WiKicker::WikiPage::ListItemNode->new;
  $list_item_node->level($level);
  $list_item_node->ordered($ordered);
  $list_item_node->add($self->_parse_inline($data));
  return $list_item_node;
}

sub _parse_table_record;
sub _split_to_cell;
sub do_table;
sub do_pre;
sub do_quote;
sub do_dl;
sub set_quote_to_stack_top;
sub set_dl_to_stack_top;
sub _make_quote_line_node;
sub _make_dl_item_node;

1;

__END__
sub _parse_table_record {
  my ($self, $text, $type) = @_;

  die "Invalid type: $type" unless $type eq '|' || $type eq ',';

  my ($line, $new_line);
  if ($text =~ m/^(.*)\n$/) {
    $line = $1;
    $new_line = "\n";
  }
  else {
    $line = $text;
    $new_line = '';
  }

  my @cells = _split_to_cell($line, $type);
  my @cell_nodes;
  for my $cell (@cells) {
    my $cell_node = WiKicker::WikiPage::TableCellNode->new;
    $cell_node->separator($type);
    my ($left, $body, $right) = _trim($cell);
    $cell_node->left($left);
    my @inline = $self->_parse_inline($body);
    if (@inline) {
      $cell_node->add(@inline);
    }
    else {
      $cell_node->add(WiKicker::WikiPage::TextNode->new(''));
    }
    $cell_node->right($right);
    push(@cell_nodes, $cell_node);
  }

  my $tr_node = WiKicker::WikiPage::TableRecordNode->new;
  $tr_node->add(@cell_nodes);
  $tr_node->cr($new_line);
  return $tr_node;
}

sub _split_to_cell {
  my ($line, $type) = @_;

  if ($line eq '') {
    return '';
  }

  my @result;
  my $pos = 0;
  my $next_pos;
  my $separator_length = length($type);
  while (($next_pos = index($line, $type, $pos)) != -1) {
    #warn "\n$pos:$next_pos";
    push(@result, substr($line, $pos, $next_pos - $pos));
    $pos = $next_pos + $separator_length;
  }
  push(@result, substr($line, $pos));

  return @result;
}

sub do_table {
  my ($self, $data, $type) = @_;

  WiKicker::Loader::load_module('WiKicker::WikiPage::TableNode');
  WiKicker::Loader::load_module('WiKicker::WikiPage::TableRecordNode');
  WiKicker::Loader::load_module('WiKicker::WikiPage::TableCellNode');

  if (ref $self->current_node ne 'WiKicker::WikiPage::TableNode') {
    $self->clear_node_stack;
    $self->push_node(WiKicker::WikiPage::TableNode->new);
  }

  $self->current_node->add($self->_parse_table_record($data, $type));
}

sub do_pre {
  my ($self, $data) = @_;

  if ($self->current_node->isa('WiKicker::WikiPage::EmptyLinesNode')) {
    my @nodes = $self->document_node->nodes;
    if (@nodes >= 2) {
      if ($nodes[-2]->isa('WiKicker::WikiPage::PreNode')) {
        my $empty_lines = $self->current_node->to_string;
        $self->document_node->remove_last_node;
        $self->clear_node_stack;
        push(@{$self->{current_path}}, $nodes[-2]);
        $self->current_node->add($empty_lines);
        $self->current_node->add(' ' . $data);
        return 1;
      }
    }
  }

  if (ref $self->current_node ne 'WiKicker::WikiPage::PreNode') {
    $self->clear_node_stack;
    $self->push_node(WiKicker::WikiPage::PreNode->new);
  }

  $self->current_node->add(' ' . $data);
}

sub do_quote {
  my ($self, $level, $data) = @_;

  #WiKicker::Loader::load_module('WiKicker::WikiPage::QuoteNode');
  #WiKicker::Loader::load_module('WiKicker::WikiPage::QuoteLineNode');
  $self->set_quote_to_stack_top($level);
  $self->current_node->add($self->_make_quote_line_node($level, $data));
}

sub do_dl {
  my ($self, $level, $data) = @_;

  WiKicker::Loader::load_module('WiKicker::WikiPage::DlNode');
  WiKicker::Loader::load_module('WiKicker::WikiPage::DtNode');
  WiKicker::Loader::load_module('WiKicker::WikiPage::DdNode');
  $self->set_dl_to_stack_top($level);
  $self->current_node->add($self->_make_dl_item_node($level, $data));
}

sub set_quote_to_stack_top {
  my ($self, $level) = @_;

  while (1) {
    if (ref $self->current_node eq 'WiKicker::WikiPage::DocumentNode') {
      for (my $i = 1; $i <= $level; $i++) {
        my $quote_node = WiKicker::WikiPage::QuoteNode->new;
        $quote_node->level($i);
        $self->push_node($quote_node);
      }
      return;
    }
    if (ref $self->current_node eq 'WiKicker::WikiPage::QuoteNode') {
      my $current_level = $self->current_node->level;
      if ($current_level == $level) {
        return;
      }
      elsif ($current_level > $level) {
        $self->pop_node;
        next;
      }
      else {
        my $quote_node = WiKicker::WikiPage::QuoteNode->new;
        $quote_node->level($current_level + 1);
        $self->push_node($quote_node);
        next;
      }
    }
    $self->clear_node_stack;
  }
}

sub set_dl_to_stack_top {
  my ($self, $level) = @_;

  while (1) {
    if (ref $self->current_node eq 'WiKicker::WikiPage::DocumentNode') {
      my $dl_node = WiKicker::WikiPage::DlNode->new;
      $dl_node->level(1);
      $self->push_node($dl_node);
      next;
    }
    elsif (ref $self->current_node eq 'WiKicker::WikiPage::DlNode') {
      my $current_level = $self->current_node->level;
      if ($current_level == $level) {
        return;
      }
      elsif ($current_level > $level) {
        $self->pop_node;
        next;
      }
      else {
        my @nodes = $self->current_node->nodes;
        if (@nodes > 0 && ref $nodes[-1] eq 'WiKicker::WikiPage::DdNode') {
          my $dl_node = WiKicker::WikiPage::DlNode->new;
          $dl_node->level($current_level + 1);
          $nodes[-1]->add($dl_node);
          push(@{$self->{current_path}}, $dl_node);
          next;
        }
        else {
          my $dd_node = WiKicker::WikiPage::DdNode->new;
          $dd_node->level($current_level);
          $dd_node->dummy(1);
          $self->current_node->add($dd_node);

          my $dl_node = WiKicker::WikiPage::DlNode->new;
          $dl_node->level($current_level + 1);

          $dd_node->add($dl_node);
          push(@{$self->{current_path}}, $dl_node);
          next;
        }
      }
    }
    $self->clear_node_stack;
  }
}

sub _make_quote_line_node {
  my ($self, $level, $data) = @_;
  my $quote_line_node = WiKicker::WikiPage::QuoteLineNode->new;
  $quote_line_node->level($level);
  $quote_line_node->add($self->_parse_inline($data));
  return $quote_line_node;
}

sub _make_dl_item_node {
  my ($self, $level, $data) = @_;

  $data =~ /^([^:]*)/;
  my $dt = $1;
  my $dt_node = WiKicker::WikiPage::DtNode->new;
  $dt_node->level($level);
  $dt_node->add($self->_parse_inline($dt));
  if (length($dt) == length($data)) {
    return $dt_node;
  }
  my $dd = substr($data, length($dt)+1);
  my $dd_node = WiKicker::WikiPage::DdNode->new;
  $dd_node->level($level);
  $dd_node->add($self->_parse_inline($dd));
  return ($dt_node, $dd_node);
}
