#
# WiKicker
#
#  HtmlFragmentVisitor.pm
#
#
# Copyright (C) 2003,2004 WATANABE Yoshimasa <naney@naney.org>
#
# This file is part of the WiKicker.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc., 59
# Temple Place, Suite 330, Boston, MA 02111-1307 USA
#
#
# $Id: HtmlFragmentVisitor.pm,v 1.42 2004/08/03 12:55:52 naney Exp $
#
#

package WiKicker::WikiPage::HtmlFragmentVisitor;

use strict;
use WiKicker::URI;
use WiKicker::HTML;
use WiKicker::WikiPage::Visitor;
use AutoLoader 'AUTOLOAD';
use Carp;

use vars qw(@ISA);
@ISA = qw(WiKicker::WikiPage::Visitor);

sub new {
  my ($class, %param) = @_;

  my $self = bless { content => '',
                     comment_number => -1,
                     footnote_number => 1,
                     footnote => [],
                   }, $class;

  for my $key (keys %param) {
    $self->{$key} = $param{$key};
  }

  return $self;
}

sub DESTROY {
  my $self = shift;

  #-IFDEBUG: WiKicker::cycle_check($self);
  #-IFDEBUG: WiKicker::total_size("HtmlFragmentVisitor(before DESTROY)", $self);
}

sub flush_closure {
  my ($self, $value) = @_;

  if (defined $value) {
    $self->{flush_closure} = $value;
  }
  return $self->{flush_closure};
}

sub in_inline {
  my ($self, $value) = @_;

  if (defined $value) {
    $self->env->in_inline($value);
  }
  return $self->env->in_inline;
}

sub form_charset {
  my ($self, $value) = @_;

  if (defined $value) {
    $self->{form_charset} = $value;
  }
  return $self->{form_charset};
}

sub no_edit_section {
  my ($self, $value) = @_;

  if (defined $value) {
    $self->{no_edit_section} = $value;
  }
  return $self->{no_edit_section};
}

sub no_automatic_link {
  my ($self, $value) = @_;

  if (defined $value) {
    $self->{no_automatic_link} = $value;
  }
  return $self->{no_automatic_link};
}

sub word_extractor {
  my ($self, $value) = @_;

  if (defined $value) {
    $self->{word_extractor} = $value;
  }
  return $self->{word_extractor};
}

sub env {
  my ($self, $value) = @_;

  if (defined $value) {
    $self->{env} = $value;
  }
  return $self->{env};
}

sub wri_factory {
  my ($self, $value) = @_;

  if (defined $value) {
    $self->{wri_factory} = $value;
  }
  return $self->{wri_factory};
}

sub call_flush_closure {
  my $self = shift;

  if (defined $self->flush_closure) {
    &{$self->flush_closure}($self->content);
    $self->{content} = '';
  }
}

sub _initialize {
  my $self = shift;

  croak 'no env' unless defined $self->env;

  $self->_initialize_word_extractor;

  $self->{anchor} = 0;
  $self->{section} = 0;
  $self->in_inline(0);
}

sub _initialize_word_extractor {
  my $self = shift;

  unless ($self->no_automatic_link) {
    $self->_initialize_word_extractor_imp;
  }
}

sub visit_DocumentNode {
  my ($self, $document_node) = @_;

  $self->_initialize;

  my $nodes = $document_node->child_nodes;
  if (scalar(@$nodes) > 0) {
    my $node = $nodes->[0];
    unless ($node->isa('WiKicker::WikiPage::HeadingNode')) {
      $self->_start_article;
      $self->_start_article_body;
      $self->{not_first_article} = 1;
    }
  }

  for my $node (@$nodes) {
    $node->accept($self);
    $self->call_flush_closure;
  }
  $self->write_footnote;
  $self->_add_edit_section_div($self->{section});

  if ($self->{not_first_article}) {
    $self->_end_section;
    $self->_end_article_body;
    $self->_end_article;
  }
  $self->call_flush_closure;
}

sub _start_article {
  my $self = shift;

  # for hook
}

sub _end_article {
  my $self = shift;

  #for hook
}

sub _start_article_body {
  my $self = shift;

  # for hook
}

sub _end_article_body {
  my $self = shift;

  # for hook
}

sub _start_section {
  my $self = shift;

  # for hook
}

sub _end_section {
  my $self = shift;

  # for hook
}

sub visit_ParagraphNode {
  my ($self, $paragraph_node) = @_;

  my $nodes = $paragraph_node->child_nodes;
  if ($paragraph_node->is_index_or_form_only) {
    $self->in_inline(0);
    $nodes->[0]->accept($self);
    return;
  }

  $self->{content} .= '<p>';
  $self->in_inline(1);
  for my $node (@$nodes) {
    $node->accept($self);
  }
  $self->in_inline(0);
  $self->{content} .= "</p>\n";
}



sub visit_HeadingNode {
  my ($self, $heading_node) = @_;

  $self->_visit_HeadingNode_hook($heading_node);

  if ($heading_node->level <= 2) {
    $self->write_footnote;
    $self->_add_edit_section_div($self->{section});
    $self->{section}++;
  }
  $self->{anchor}++;

  if ($heading_node->level == 1) {
    if ($self->{not_first_article}) {
      $self->_end_section;
      $self->_end_article_body;
      $self->_end_article;
    }
    $self->{not_first_article} = 1;
    $self->_start_article;
  }
  elsif ($heading_node->level == 2) {
    $self->_end_section;
    $self->_start_section;
  }

  my $level = $heading_node->level + $self->_heading_level_adjust;

  my $anchor_name = $self->{anchor};
  my $heading_prefix_mark = $self->heading_prefix_mark($heading_node->level);
  $self->{content} .= qq(<h$level>$heading_prefix_mark<a name="h$anchor_name"></a>);
  $self->in_inline(1);
  for my $node (@{$heading_node->child_nodes}) {
    $node->accept($self);
  }
  chomp $self->{content};
  $self->in_inline(0);
  $self->{content} .= qq(</h$level>\n);

  if ($heading_node->level == 1) {
    $self->_start_article_body;
  }
}

# for override
sub _visit_HeadingNode_hook {}

# for override
sub _heading_level_adjust { 1 }

# for override
sub heading_prefix_mark {
  my ($self, $level) = @_;

  return '';
}


sub visit_TextNode {
  my ($self, $text) = @_;

  if ($self->no_automatic_link) {
    $self->{content} .= WiKicker::HTML::escape($text->to_string);
  }
  else {
    if ($self->{automatic_link_type} eq 'single') {
      $self->_visit_TextNode_for_automatic_link($text);
    }
    else {
      $self->_visit_TextNode_for_automatic_link_multi($text);
    }
  }
}


sub visit_VerbatimNode {
  my ($self, $node) = @_;

  $self->{content} .= WiKicker::HTML::escape($node->text);
}

sub visit_WriNode {
  my ($self, $wri_node) = @_;

  $self->_add_wri_content($wri_node->wri);
}

sub visit_BracketNameNode {
  my ($self, $bracket_name_node) = @_;

  if ($bracket_name_node->is_empty_bracket_name) {
    $self->{content} .= $bracket_name_node->original;
  }
  else {
    $self->_add_wri_content($bracket_name_node->wri,
                            $bracket_name_node->content);
  }
}

sub visit_HrNode {
  my ($self, $hr_node) = @_;

  $self->{content} .= "<hr>\n";
}

sub visit_EmptyLinesNode {
  my ($self, $empty_lines_node) = @_;

  # nothing to do
}

sub visit_ListNode {
  my ($self, $list_node) = @_;

  my $element;
  if ($list_node->ordered) {
    $element = 'ol';
  }
  else {
    $element = 'ul';
  }

  $self->{content} .= '    ' x ($list_node->level - 1);
  $self->{content} .= qq(<$element>\n);

  for my $node (@{$list_node->child_nodes}) {
    $node->accept($self);
  }

  $self->{content} .= '    ' x ($list_node->level - 1);
  $self->{content} .= "</$element>\n";
  $self->in_inline(0);
}

sub _visit_normal_node {
  my ($self, $node, $element_name) = @_;

  $self->{content} .= "<$element_name>";
  for my $child (@{$node->child_nodes}) {
    $child->accept($self);
  }
  $self->{content} .= "</$element_name>";
}

sub _visit_normal_chomp_node {
  my ($self, $node, $element_name) = @_;

  $self->{content} .= "<$element_name>";
  for my $child (@{$node->child_nodes}) {
    $child->accept($self);
  }
  chomp $self->{content};
  $self->{content} .= "</$element_name>";
}

sub visit_ListItemNode {
  my ($self, $node) = @_;

  $self->{content} .= '  ';
  $self->{content} .= '    ' x ($node->level - 1);
  $self->in_inline(0);
  $self->_visit_normal_chomp_node($node, 'li');
  $self->{content} .= "\n";

  $self->call_flush_closure; # for long list
}


sub _add_wri_content {
  my ($self, $wri, $content) = @_;

  $self->{content} .= $wri->to_html($content, $self->env);
}

sub content {
  return shift->{content};
}

sub _add_edit_section_div {
  my ($self, $section_number) = @_;

  unless ($self->no_edit_section) {
    my $link = $self->_section_edit_link($section_number);

    $self->{content}
      .= qq(<div class="section-toolbar">[$link]</div>\n);
  }
}

sub _section_edit_uri {
  my ($self, $section_number) = @_;

  return $self->env->link_maker->edit_uri($self->env->page_name)
    . '&amp;section=' . $section_number;
}

sub _section_edit_link {
  my ($self, $section_number) = @_;

  my $section_url = $self->_section_edit_uri($section_number);
  my $title = $self->env->localizer->maketext('edit this section');
  my $text = $self->env->localizer->maketext('edit');
  return qq(<a href="$section_url" title="$title">$text</a>);
}


sub visit_PreNode;
sub visit_QuoteNode;
sub visit_DlNode;
sub visit_DtNode;
sub visit_DdNode;
sub visit_QuoteLineNode;
sub add_footnote;
sub write_footnote;
sub _visit_TextNode_for_automatic_link;
sub _visit_TextNode_for_automatic_link_multi;
sub _initialize_word_extractor_imp;
sub visit_EmphasisNode;
sub visit_InlineBlockNode;
sub visit_TableNode;
sub visit_TableRecordNode;
sub visit_TableCellNode;
sub _automatic_linked;

1;

__END__
sub visit_PreNode {
  my ($self, $pre_node) = @_;

  $self->{content} .= '<pre>';
  $self->in_inline(1);
  $self->{content} .= WiKicker::HTML::escape($pre_node->to_string);
  $self->in_inline(0);
  $self->{content} .= "</pre>\n";
}

sub visit_QuoteNode {
  my ($self, $node) = @_;

  my $in_p = 0;

  $self->{content} .= '<blockquote>';
  for my $child ($node->nodes) {
    if ($child->isa('WiKicker::WikiPage::QuoteNode')) {
      if ($in_p) {
        $self->{content} .= '</p>';
        $in_p = 0;
        $self->in_inline(0);
      }
    }
    else {
      unless ($in_p) {
        $self->{content} .= '<p>';
        $in_p = 1;
        $self->in_inline(1);
      }
    }
    $child->accept($self);
  }
  if ($in_p) {
    $self->in_inline(0);
    $self->{content} .= '</p>';
  }
  $self->{content} .= "</blockquote>\n";
}

sub visit_QuoteLineNode {
  my ($self, $line_node) = @_;
  for my $node ($line_node->nodes) {
    $node->accept($self);
  }
}

sub visit_DlNode {
  my ($self, $dl_node) = @_;

  $self->{content} .= '    ' x ($dl_node->level - 1);
  $self->{content} .= qq(<dl>\n);
  for my $node ($dl_node->nodes) {
    $node->accept($self);
  }
  $self->{content} .= '    ' x ($dl_node->level - 1);
  $self->{content} .= "</dl>\n";
}

sub visit_DtNode {
  my ($self, $node) = @_;

  my $child_string = $node->get_child_string;
  unless ($child_string =~ /^\s*$/) {
    $self->{content} .= '  ';
    $self->{content} .= '    ' x ($node->level - 1);
    $self->in_inline(1);
    $self->_visit_normal_chomp_node($node, 'dt');
    $self->in_inline(0);
    $self->{content} .= "\n";
  }
}

sub visit_DdNode {
  my ($self, $node) = @_;

  $self->{content} .= '  ';
  $self->{content} .= '    ' x ($node->level - 1);
  $self->in_inline(0);
  $self->_visit_normal_chomp_node($node, 'dd');
  $self->{content} .= "\n";
}

sub add_footnote {
  my ($self, $inline_block_node) = @_;

  my $n = $self->{footnote_number}++;
  $self->{content} .= qq(<span class="footnote"><a name="fm$n" href="#f$n">*$n</a></span>);
  push(@{$self->{footnote}}, [$n, $inline_block_node]);
}

sub write_footnote {
  my $self = shift;

  my $footnote_list = $self->{footnote};
  unless (scalar(@$footnote_list)) {
    return;
  }
  $self->{content} .= qq(<div class="footnote">);
  for my $item (@$footnote_list) {
    my $n = $item->[0];
    $self->{content} .= qq(<p class="footnote"><a name="f$n" href="#fm$n">*$n</a>);
    my $inline_block_node = $item->[1];
    for my $node ($inline_block_node->nodes) {
      $node->accept($self);
    }
    $self->{content} .= qq(</p>);
  }
  $self->{content} .= qq(</div>);
  $self->{footnote} = [];
}

sub _automatic_linked {
  my ($self, $page_name) = @_;
  # for override
}

sub _visit_TextNode_for_automatic_link {
  my ($self, $text) = @_;

  my $splited = $self->word_extractor->split_into_words($text->to_string);
  for my $item (@$splited) {
    if ($item->[0] == 0) {
      $self->{content} .= WiKicker::HTML::escape($item->[1]);
    }
    else {
      my $page_name = $item->[1];
      $page_name = $item->[2] if defined $item->[2];
      $self->_automatic_linked($page_name);
      my $href = $self->env->link_maker->make_page_read_href($page_name);
      my $link_text = WiKicker::HTML::escape($item->[1]);
      $self->{content} .= qq(<a class="auto" href="$href">$link_text</a>);
    }
  }
}

sub _visit_TextNode_for_automatic_link_multi {
  my ($self, $text) = @_;

  my $splited = $self->word_extractor->split_into_words($text->to_string);
  for my $item (@$splited) {
    if ($item->[0] == 0) {
      $self->{content} .= WiKicker::HTML::escape($item->[1]);
    }
    else {
      my $page_name = $item->[1];
      $page_name = $item->[2] if defined $item->[2];
      $self->_automatic_linked($page_name);
      my $link_text = WiKicker::HTML::escape($item->[1]);
      if ($item->[0] == 1) {
        my $href = $self->env->link_maker->make_page_read_href($page_name);
        $self->{content} .= qq(<a class="auto" href="$href">$link_text</a>);
      }
      else {
        my $href = $self->env->resolver
          ->get_interwiki_link_url($self->{extraset_scheme}->[$item->[0] - 2],
                                   $page_name);
        if ($href ne '') {
          $self->{content} .= qq(<a class="auto" href="$href">$link_text</a>);
        }
        else {
          $self->{content} .= $link_text;
        }
      }
    }
  }
}

sub _initialize_word_extractor_imp {
  my $self = shift;

  my $extraset = $self->env->extraset;
  if (defined  $extraset && scalar(@$extraset) >= 1) {
    eval "use WiKicker::Text::MultiSetWordExtractor";
    if ($@) {
      croak "Can't load WiKicker::Text::MultiSetWordExtractor: $@";
    }
    $self->word_extractor(WiKicker::Text::MultiSetWordExtractor->new);
    my $words_list_list = [];
    if ($self->env->page_name_list) {
      push(@$words_list_list, $self->env->page_name_list->list);
    } else {
      push(@$words_list_list, []);
    }
    push(@$words_list_list, map {$_->{words}} @$extraset);
    $self->word_extractor->set_words_list_list($words_list_list);

    my @schemes = map {$_->{scheme}} @$extraset;
    $self->{extraset_scheme} = \@schemes;
    $self->{automatic_link_type} = 'multi';
  } else {
    eval "use WiKicker::Text::WordExtractor";
    if ($@) {
      croak "Can't load WiKicker::Text::WordExtractor: $@";
    }
    $self->word_extractor(WiKicker::Text::WordExtractor->new);
    if ($self->env->page_name_list) {
      $self->word_extractor->set_words($self->env->page_name_list->list);
    }
    $self->{automatic_link_type} = 'single';
  }
}

sub visit_EmphasisNode {
  my ($self, $emphasis_node) = @_;
  $self->{content} .= '<em>';
  for my $node (@{$emphasis_node->child_nodes}) {
    $node->accept($self);
  }
  $self->{content} .= '</em>';
}

sub visit_InlineBlockNode {
  my ($self, $inline_block_node) = @_;

  my $scheme = $inline_block_node->scheme;
  if ($scheme ne '') {
    if ($scheme eq 'fn') {
      $self->add_footnote($inline_block_node);
    }
    else {
      $self->{content} .= '{{' . $scheme . ':';
      for my $node ($inline_block_node->nodes) {
        $node->accept($self);
      }
      $self->{content} .= '}}';
    }
  }
  else {
    $self->{content} .= '{{';
    for my $node ($inline_block_node->nodes) {
      $node->accept($self);
    }
    $self->{content} .= '}}';
  }
}

sub visit_TableNode {
  my ($self, $node) = @_;

  $self->_visit_normal_node($node, 'table');
  $self->{content} .= "\n";
}

sub visit_TableRecordNode {
  my ($self, $record_node) = @_;


  $self->{content} .= qq(<tr>);

  for my $node ($record_node->nodes) {
    $node->accept($self);
  }
  $self->{content} .= qq(</tr>\n);
}

sub visit_TableCellNode {
  my ($self, $cell_node) = @_;

  my $align = $cell_node->align;
  my $align_attribute = '';
  if ($align eq 'center') {
    $align_attribute = qq( align="center");
  }
  elsif ($align eq 'right') {
    $align_attribute = qq( align="right");
  }

  $self->{content} .= qq(<td$align_attribute>);
  $self->in_inline(0);
  for my $node ($cell_node->nodes) {
    $node->accept($self);
  }
  $self->{content} .= qq(</td>);
}
